# initialise the module folder
__all__ = [
    'create_disk', 'get_cylinder', 'imdiff', 'imscroll', 'profile_points', 'pifa2nii', 'nii2pifa']

from .gen import create_disk, get_cylinder, imdiff, imscroll, profile_points
from .signa import nii2pifa, pifa2nii



#!/usr/bin/env python
"""initialise the NiftyPET NIMPA package"""
__author__ = "Pawel J. Markiewicz", "Casper O. da Costa-Luis"
__copyright__ = "Copyright 2021"
# version detector. Precedence: installed dist, git, 'UNKNOWN'
try:
    from ._dist_ver import __version__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version(root="../..", relative_to=__file__)
    except (ImportError, LookupError):
        __version__ = "UNKNOWN"
__all__ = [
    # gpu utils
    'cs', 'dev_info', 'gpuinfo',
    # utils
    'LOG_FORMAT', 'LogHandler',
    # config
    'path_resources', 'resources', 'cmake_prefix',
    # numcu
    'add', 'div', 'mul',
    # improc
    'conv_separable', 'isub', 'nlm', 'aff_dist', 'centre_mass_rel',
    # core
    'create_disk', 'get_cylinder', 'imdiff', 'imscroll', 'profile_points', 'imtrimup',
    'affine_fsl', 'affine_dipy', 'affine_niftyreg',
    'array2nii', 'bias_field_correction',
    'centre_mass_img', 'centre_mass_corr', 'coreg_spm', 'coreg_vinci',
    'create_dir', 'create_mask', 'ct2mu',
    'dcm2im', 'dcm2nii', 'dcmanonym', 'dcminfo', 'dcmsort', 'isdcm', 'dcmdir',
    'dice_coeff', 'dice_coeff_multiclass', 'fwhm2sig', 'getmgh', 'getnii', 'mgh2nii',
    'getnii_descr', 'im_cut', 'imfill', 'imsmooth', 'iyang', 'motion_reg', 'nii_gzip',
    'nii_modify', 'nii_ugzip', 'niisort', 'orientnii', 'pet2pet_rigid', 'pick_t1w',
    'psf_gaussian', 'psf_measured', 'pvc_iyang', 'realign_mltp_spm', 'resample_fsl',
    'resample_mltp_spm', 'resample_niftyreg', 'resample_spm', 'resample_vinci', 'resample_dipy',
    'time_stamp', 'rem_chars',
    # Signa
    'pifa2nii', 'nii2pifa',
    # ACR
    'acr'
    # 'get_params', 'get_paths', 'extract_reso_part', 'sampling_masks'
    # 'create_mumap_core', 'create_nac_core', 'create_reso', 'create_sampl_reso', 'create_sampl',
    # 'standard_analysis', 'estimate_fwhm'
    ] # yapf: disable

from os import fspath

try:          # py<3.9
    import importlib_resources as iresources
except ImportError:
    from importlib import resources as iresources

try:
    from numcu import add, div, mul
except ImportError:
    pass

from niftypet.ninst import cudasetup as cs
from niftypet.ninst.dinf import dev_info, gpuinfo
from niftypet.ninst.tools import LOG_FORMAT, LogHandler, path_resources, resources

from . import acr
from .img import create_disk, get_cylinder, imdiff, imscroll, nii2pifa, pifa2nii, profile_points
from .prc import imtrimup  # for backward compatibility
from .prc import (
    aff_dist,
    affine_dipy,
    affine_fsl,
    affine_niftyreg,
    array2nii,
    bias_field_correction,
    centre_mass_corr,
    centre_mass_img,
    centre_mass_rel,
    conv_separable,
    coreg_spm,
    coreg_vinci,
    create_dir,
    create_mask,
    ct2mu,
    dcm2im,
    dcm2nii,
    dcmanonym,
    dcmdir,
    dcminfo,
    dcmsort,
    dice_coeff,
    dice_coeff_multiclass,
    fwhm2sig,
    getmgh,
    getnii,
    getnii_descr,
    im_cut,
    imfill,
    imsmooth,
    isdcm,
    isub,
    iyang,
    mgh2nii,
    motion_reg,
    nii_gzip,
    nii_modify,
    nii_ugzip,
    niisort,
    nlm,
    orientnii,
    pet2pet_rigid,
    pick_t1w,
    psf_gaussian,
    psf_measured,
    pvc_iyang,
    realign_mltp_spm,
    rem_chars,
    resample_dipy,
    resample_fsl,
    resample_mltp_spm,
    resample_niftyreg,
    resample_spm,
    resample_vinci,
    time_stamp,
)

# for use in `cmake -DCMAKE_PREFIX_PATH=...`
cmake_prefix = fspath(iresources.files("niftypet.nimpa").resolve() / "cmake")



"""ACR/Jaszczak PET phantom templates"""
__author__ = ("Pawel Markiewicz", "Casper da Costa-Luis")
__copyright__ = "Copyright 2021-23"
import os
from pathlib import Path

import imageio
import numpy as np
import scipy.ndimage as ndi

from niftypet import nimpa

# MU-MAPs


def create_mumap_core(Cntd, return_raw=False):
    """ACR core mu-map generation without the resolution rods"""
    if all(os.path.isfile(Cntd['out'][f]) for f in ('facrmu', 'facrdmu')):
        print('i> found already generated mu-map at:\n{}\nand\n{}'.format(
            Cntd['out']['facrmu'], Cntd['out']['facrdmu']))
        return None

    if 'fnacup' in Cntd and Path(Cntd['fnacup']).is_file():
        imupd = nimpa.getnii(Cntd['fnacup'], output='all')
    elif 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        imupd = nimpa.getnii(Cntd['fqntup'], output='all')
    else:
        raise ValueError('Upscaled and trimmed ACR PET image cannot be found')

    # > main cap
    mcap0_png = imageio.imread(Cntd['fmcap0'])
    mcap1_png = imageio.imread(Cntd['fmcap1'])
    mcap2_png = imageio.imread(Cntd['fmcap2'])

    # > bone and other screws
    bscrw_png = imageio.imread(Cntd['fbscrw'])
    scrws_png = imageio.imread(Cntd['fscrws'])

    # > lid
    lid0_png = imageio.imread(Cntd['flid0'])
    lid1_png = imageio.imread(Cntd['flid1'])

    # > cylindrical inserts attached to the lid
    insrt_png = imageio.imread(Cntd['finsrt'])
    binsrt_png = imageio.imread(Cntd['fbinsrt'])

    # > the main phantom container with the bottom
    main_png = imageio.imread(Cntd['fmain'])
    bttm_png = imageio.imread(Cntd['fbttm'])

    # > convert image slices to mu-map values
    # > main cap, part 0 (A)
    mcap0 = np.float32((mcap0_png[..., 0] == Cntd['png_scrws']) * Cntd['mu_screw'])
    mcap0 = ndi.zoom(mcap0, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(mcap0, cmap='bone', vmin=0.07, vmax=0.14)
    # > main cap, part 1 (B)
    mcap1 = np.float32((mcap1_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']) + np.float32(
        (mcap1_png[..., 0] == Cntd['png_scrws']) * Cntd['mu_screw'])
    mcap1 = ndi.zoom(mcap1, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(mcap1, cmap='bone', vmin=0.07, vmax=0.14)
    # > main cap, part 2 (C)
    mcap2 = np.float32((mcap2_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']) + np.float32(
        (mcap2_png[..., 0] == Cntd['png_water']) * Cntd['mu_water'])
    # prnt = mcap2
    mcap2 = ndi.zoom(mcap2, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(mcap2, cmap='bone', vmin=0.07, vmax=0.14)
    # > normal screws
    ascrws = np.float32((scrws_png[..., 0] == Cntd['png_scrws']) * Cntd['mu_screw'])
    # prnt += ascrws
    ascrws = ndi.zoom(ascrws, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(ascrws, cmap='bone', vmin=0.07, vmax=0.14)
    # > bone screw
    bscrws = np.float32((bscrw_png[..., 0] == Cntd['png_scrws']) * Cntd['mu_screw'])
    # prnt += bscrws
    bscrws = ndi.zoom(bscrws, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(ascrws+bscrws, cmap='bone', vmin=0.07, vmax=0.14)
    # matshow(prnt, cmap='bone', vmin=0.07, vmax=0.14)

    # > multiple slices according to the thickness of different part (axially)

    # > main cap with the three parts
    k = int(np.round(Cntd['k_mcapA'] / Cntd['vxsz']))
    mcapA = np.repeat(mcap0[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_mcapB'] / Cntd['vxsz']))
    mcapB = np.repeat(mcap1[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_mcapC'] / Cntd['vxsz']))
    tmp = (ascrws + mcap1)
    mcapC = np.repeat(tmp[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_mcapD'] / Cntd['vxsz']))
    tmp = (ascrws + mcap2 + bscrws)
    mcapD = np.repeat(tmp[None, ...], k, axis=0)

    # matshow(mcapD[0,...], cmap='bone')

    # > lid in two parts
    lid0 = np.float32((lid0_png[..., 0] == Cntd['png_scrws']) * Cntd['mu_screw']) + np.float32(
        (lid0_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']) + np.float32(
            (lid0_png[..., 0] == Cntd['png_water']) * Cntd['mu_water'])
    lid0 = ndi.zoom(lid0, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(lid0, cmap='bone', vmin=0.07, vmax=0.14)
    lid1 = np.float32((lid1_png[..., 0] == Cntd['png_scrws']) * Cntd['mu_screw']) + np.float32(
        (lid1_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']) + np.float32(
            (lid1_png[..., 0] == Cntd['png_water']) * Cntd['mu_water'])
    lid1 = ndi.zoom(lid1, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(lid1, cmap='bone', vmin=0.07, vmax=0.14)

    k = int(np.round(Cntd['k_lidA'] / Cntd['vxsz']))
    lidA = np.repeat(lid0[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_lidB'] / Cntd['vxsz']))
    lidB = np.repeat(lid1[None, ...], k, axis=0)

    # > cylindrical inserts
    insrt = np.float32((insrt_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']) + np.float32(
        (insrt_png[..., 0] == Cntd['png_boney']) * Cntd['mu_bone']) + np.float32(
            (insrt_png[..., 0] == Cntd['png_water']) * Cntd['mu_water'])
    insrt = ndi.zoom(insrt, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(insrt, cmap='bone', vmin=0.07, vmax=0.14)
    # > bone insert
    insrtb = np.float32((binsrt_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']) + np.float32(
        (binsrt_png[..., 0] == Cntd['png_water']) * Cntd['mu_water'])
    insrtb = ndi.zoom(insrtb, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(insrtb, cmap='bone', vmin=0.07, vmax=0.14)

    k = int(np.round(Cntd['k_insrtA'] / Cntd['vxsz']))
    insrtA = np.repeat(insrt[None, ...], k, axis=0)
    '''
    matshow(insrtA[0,...], cmap='bone', vmin=0.08, vmax=0.12)
    '''

    k = int(np.round(Cntd['k_insrtB'] / Cntd['vxsz']))
    insrtB = np.repeat(insrtb[None, ...], k, axis=0)

    # > main compartment
    main = np.float32((main_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']) + np.float32(
        (main_png[..., 0] == Cntd['png_water']) * Cntd['mu_water'])
    main = ndi.zoom(main, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(main, cmap='bone', vmin=0.07, vmax=0.14)
    # > bottom of the phantom
    bttm = np.float32((bttm_png[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx'])
    bttm = ndi.zoom(bttm, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    # matshow(bttm, cmap='bone', vmin=0.07, vmax=0.14)

    k = int(np.round(Cntd['k_contA'] / Cntd['vxsz']))
    contA = np.repeat(main[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_contB'] / Cntd['vxsz']))
    contB = np.repeat(bttm[None, ...], k, axis=0)

    # > ASSEMBLE ALL PARTS
    # > whole container - concatenate all the bits together
    acr = np.concatenate((mcapA, mcapB, mcapC, mcapD, lidA, lidB, insrtA, insrtB, contA, contB),
                         axis=0)

    # > pad the image to make the dims even
    acr = np.pad(acr, ((0, 0), (Cntd['dpad'], Cntd['dpad'] + 1), (Cntd['dpad'], Cntd['dpad'] + 1)),
                 constant_values=((0, 0), (0, 0), (0, 0)))

    # > make it up-side down (as it is scanned)
    acr = acr[::-1, :, :]

    # > get the affine and the save to NIfTI
    imxys = acr.shape[2]
    imzys = acr.shape[0]
    affine = np.array([[-Cntd['vxsz'], 0., 0., .5 * imxys * Cntd['vxsz']],
                       [0., Cntd['vxsz'], 0., -.5 * imxys * Cntd['vxsz']],
                       [0., 0., Cntd['vxsz'], -.5 * imzys * Cntd['vxsz']], [0., 0., 0., 1.]])

    nimpa.array2nii(
        acr, affine, Cntd['out']['facrmu'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])
    '''
    nimpa.imscroll(facrmu, view='s', vmin=0.08, vmax=0.12, cmap='bone')
    nimpa.imscroll(acr, view='s', vmin=0.07, vmax=0.14, cmap='bone')
    matshow(acr[...,600])
    matshow(imup['im'][...,300])
    '''

    # > scale down the image; used for reducing the registration time

    acrd = ndi.zoom(acr, 1 / Cntd['scld'], output=None, order=1, mode='constant')

    # > affine
    imxys = acrd.shape[2]
    imzys = acrd.shape[0]
    affined = np.array(
        [[-Cntd['vxsz'] * Cntd['scld'], 0., 0., .5 * imxys * Cntd['vxsz'] * Cntd['scld']],
         [0., Cntd['vxsz'] * Cntd['scld'], 0., -.5 * imxys * Cntd['vxsz'] * Cntd['scld']],
         [0., 0., Cntd['vxsz'] * Cntd['scld'], -.5 * imzys * Cntd['vxsz'] * Cntd['scld']],
         [0., 0., 0., 1.]])

    nimpa.array2nii(
        acrd, affined, Cntd['out']['facrdmu'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    # nimpa.imscroll(acrd, view='s', vmin=0.08, vmax=0.12, cmap='bone')
    # del insrt, insrtb, insrtA, insrtB, contA, contB, main
    # del lidA, lidB, lid0, lid1, mcapA, mcapB, mcapC, mcapD

    # facr = os.path.join(os.path.dirname(imup['fim']), 'acr.npz')
    # np.savez(facr, acrd=acrd, affined=affined)

    # > PRINT:
    '''
    p = np.load(os.path.join(os.path.dirname(imup['fim']), 'print-resw.npy'))
    acr[p>0] = p[p>0]
    np.save(os.path.join(os.path.dirname(imup['fim']), 'print-acr.npy'), acr)
    a = np.load(os.path.join(os.path.dirname(imup['fim']), 'print-acr.npy'))
    nimpa.imscroll(a, view='s', vmin=0.07, vmax=0.14, cmap='bone')
    matshow(a[...,550], vmin=0.07, vmax=0.14, cmap='bone')
    matshow(a[...,430], vmin=0.07, vmax=0.14, cmap='bone')
    '''
    return {'acrd': acrd, 'acr': acr} if return_raw else None


def create_nac_core(Cntd, return_raw=False):
    """simulate NAC PET image for the main/core registration"""
    if os.path.isfile(Cntd['out']['facrad']):
        return None

    # > dictionary of upsampled and trimmed NAC pet image
    if 'fnacup' in Cntd and Path(Cntd['fnacup']).is_file():
        imupd = nimpa.getnii(Cntd['fnacup'], output='all')
    elif 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        imupd = nimpa.getnii(Cntd['fqntup'], output='all')
    else:
        raise ValueError('Upscaled and trimmed ACR PET image cannot be found')

    cap_png = imageio.imread(Cntd['fcap'])
    ins_png = imageio.imread(Cntd['fins'])
    big_png = imageio.imread(Cntd['fbig'])
    rng_png = imageio.imread(Cntd['frng'])

    cap = np.float32((cap_png[..., 0] == Cntd['png_acap']) * Cntd['abck'])
    cap = ndi.zoom(cap, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')

    ins = np.float32((ins_png[..., 0] == Cntd['png_abck']) * Cntd['abck']) + np.float32(
        (ins_png[..., 0] == Cntd['png_ains']) * Cntd['ains'])
    ins = ndi.zoom(ins, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')

    big = np.float32((big_png[..., 0] == Cntd['png_abck']) * Cntd['abck'])
    big = ndi.zoom(big, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')

    rng = np.float32((rng_png[..., 0] == Cntd['png_aedg']) * Cntd['aedg'])
    rng = ndi.zoom(rng, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    ins[rng == Cntd['aedg']] = Cntd['aedg']
    big[rng == Cntd['aedg']] = Cntd['aedg']

    # > for the gap, lower intensity
    rng[rng == Cntd['aedg']] = Cntd['aedg'] // 3
    '''
    matshow(cap, cmap='gray_r',vmax=200)
    matshow(ins, cmap='gray_r',vmax=200)
    matshow(big, cmap='gray_r',vmax=200)
    matshow(rng, cmap='gray_r',vmax=200)
    '''

    # > put unique slices together with k number representing the thickness
    # > variables with underscore act as zero-fillers
    k = int(np.round(Cntd['k_mcapA'] / Cntd['vxsz'])) + int(
        np.round(Cntd['k_mcapB'] / Cntd['vxsz'])) + int(np.round(Cntd['k_mcapC'] / Cntd['vxsz']))
    cap_ = np.zeros((k,) + cap.shape, dtype=np.float32)

    k = int(np.round(Cntd['k_mcapD'] / Cntd['vxsz']))
    capa = np.repeat(cap[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_lidA'] / Cntd['vxsz'])) + int(np.round(Cntd['k_lidB'] / Cntd['vxsz']))
    lid_ = np.zeros((k,) + cap.shape, dtype=np.float32)

    k = int(np.round(Cntd['k_insrtA'] / Cntd['vxsz']))
    insa = np.repeat(ins[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_unfrm'] / Cntd['vxsz'])) + int(
        np.round(Cntd['k_insrtB'] / Cntd['vxsz']))
    biga = np.repeat(big[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_rodsend'] / Cntd['vxsz']))
    gapa = np.repeat(rng[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_rods'] / Cntd['vxsz']))
    resa = np.repeat(big[None, ...], k, axis=0)

    k = int(np.round(Cntd['k_contB'] / Cntd['vxsz']))
    btm_ = np.zeros((k,) + cap.shape, dtype=np.float32)

    # > 3D assembly
    acra = np.concatenate((cap_, capa, lid_, insa, biga, gapa, resa, gapa, btm_), axis=0)

    # > pad the image to make the dims even
    acra = np.pad(acra,
                  ((0, 0), (Cntd['dpad'], Cntd['dpad'] + 1), (Cntd['dpad'], Cntd['dpad'] + 1)),
                  constant_values=((0, 0), (0, 0), (0, 0)))

    # > make it up-side down (as it is scanned)
    acra = acra[::-1, :, :]

    # > scale down
    acrad = ndi.zoom(acra, 1 / Cntd['scld'], output=None, order=1, mode='constant')

    # > affine
    imxys = acrad.shape[2]
    imzys = acrad.shape[0]
    affined = np.array(
        [[-Cntd['vxsz'] * Cntd['scld'], 0., 0., .5 * imxys * Cntd['vxsz'] * Cntd['scld']],
         [0., Cntd['vxsz'] * Cntd['scld'], 0., -.5 * imxys * Cntd['vxsz'] * Cntd['scld']],
         [0., 0., Cntd['vxsz'] * Cntd['scld'], -.5 * imzys * Cntd['vxsz'] * Cntd['scld']],
         [0., 0., 0., 1.]])

    nimpa.array2nii(
        acrad, affined, Cntd['out']['facrad'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    # matshow(acrad[420,...], cmap='magma')
    # matshow(imup['im'][420,...], cmap='magma')
    '''
    #PRINT
    matshow(acra[...,555], cmap='gray_r', vmax=200)
    '''

    if return_raw:
        return acrad
    else:
        return None


def create_reso(Cntd, return_raw=False):
    """Create the resolution rods for the mu-map and for registration."""
    if all(os.path.isfile(Cntd['out'][f]) for f in ('fresomu', 'fresdmu', 'fresdWmu', 'fresdQmu')):
        return None

    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        imupd = nimpa.getnii(Cntd['fqntup'], output='all')
    else:
        raise ValueError('Upscaled and trimmed ACR QNT PET image cannot be found')

    # ------------------------------------------------------
    # > form 3D digital resolution part of the phantom for registration
    # > and attenuation purposes
    renpng = imageio.imread(Cntd['frenpng'])
    respng = imageio.imread(Cntd['frespng'])

    # > these are rods mixed with water
    renWpng = imageio.imread(Cntd['frenWpng'])
    resWpng = imageio.imread(Cntd['fresWpng'])
    # ------------------------------------------------------

    ren = np.float32((renpng[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx'])
    res = np.float32((respng[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx'])
    '''
    matshow(res, cmap='bone', vmin=0.07, vmax=0.14)
    matshow(ren, cmap='bone', vmin=0.07, vmax=0.14)
    '''

    renW = (renWpng[..., 0] == Cntd['png_water']) * Cntd['mu_water'] + (
        renWpng[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']
    resW = (resWpng[..., 0] == Cntd['png_water']) * Cntd['mu_water'] + (
        resWpng[..., 0] == Cntd['png_prspx']) * Cntd['mu_prspx']
    renW = np.float32(renW)
    resW = np.float32(resW)

    # --PRINT--
    '''
    matshow(ren, cmap='gray_r', vmax= 0.14)
    matshow(res, cmap='gray_r', vmax= 0.14)
    matshow(renW, cmap='bone', vmin=0.07, vmax=0.14)
    matshow(resW, cmap='bone', vmin=0.07, vmax=0.14)
    '''
    # --END-PRINT--

    if Cntd['rods_rotate'] != 0:
        ren = ndi.rotate(ren, Cntd['rods_rotate'], reshape=False, order=Cntd['intord'],
                         mode='constant')
        res = ndi.rotate(res, Cntd['rods_rotate'], reshape=False, order=Cntd['intord'],
                         mode='constant')

        renW = ndi.rotate(renW, Cntd['rods_rotate'], reshape=False, order=Cntd['intord'],
                          mode='constant')
        resW = ndi.rotate(resW, Cntd['rods_rotate'], reshape=False, order=Cntd['intord'],
                          mode='constant')

    ren = ndi.zoom(ren, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    res = ndi.zoom(res, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')

    renW = ndi.zoom(renW, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    resW = ndi.zoom(resW, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')

    # > rod ends
    k = int(np.round(Cntd['k_rodsend'] / Cntd['vxsz']))
    ren = np.repeat(ren[None, ...], k, axis=0)
    renW = np.repeat(renW[None, ...], k, axis=0)
    '''
    matshow(ren[0,...])
    matshow(renW[0,...])
    '''

    # > rods themselves
    k = int(np.round(Cntd['k_rods'] / Cntd['vxsz']))
    res = np.repeat(res[None, ...], k, axis=0)
    resW = np.repeat(resW[None, ...], k, axis=0)

    # > create a buffer for registration purposes
    # > as a margin between the rods end and the uniform part
    bsz = Cntd['buff_rods_size']
    buff = resW.copy()
    buff = buff[:bsz, ...]
    buff[:] = 0

    # > put all together
    reso = np.concatenate((ren, res, ren, buff), axis=0)
    # reso = np.pad( reso, ((1,0), (0,1),(0,1)), constant_values=((0,0), (0,0), (0, 0)) )
    reso = np.pad(reso,
                  ((1, 0), (Cntd['dpad'], Cntd['dpad'] + 1), (Cntd['dpad'], Cntd['dpad'] + 1)),
                  constant_values=((0, 0), (0, 0), (0, 0)))

    # nimpa.imscroll(res, view='c')

    imxys = reso.shape[2]
    imzys = reso.shape[0]
    affine = np.array([[-Cntd['vxsz'], 0., 0., .5 * imxys * Cntd['vxsz']],
                       [0., Cntd['vxsz'], 0., -.5 * imxys * Cntd['vxsz']],
                       [0., 0., Cntd['vxsz'], -.5 * imzys * Cntd['vxsz']], [0., 0., 0., 1.]])

    nimpa.array2nii(
        reso,                                                            # [::-1, :, :],
        affine,
        Cntd['out']['fresomu'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)),
        flip=imupd['flip'])

    # nimpa.imscroll(fresomu, view='t')

    # > scale down
    resd = ndi.zoom(reso, 1 / Cntd['scld'], output=None, order=1, mode='constant')

    # > affine
    imxys = resd.shape[2]
    imzys = resd.shape[0]
    affined = np.array(
        [[-Cntd['vxsz'] * Cntd['scld'], 0., 0., .5 * imxys * Cntd['vxsz'] * Cntd['scld']],
         [0., Cntd['vxsz'] * Cntd['scld'], 0., -.5 * imxys * Cntd['vxsz'] * Cntd['scld']],
         [0., 0., Cntd['vxsz'] * Cntd['scld'], -.5 * imzys * Cntd['vxsz'] * Cntd['scld']],
         [0., 0., 0., 1.]])

    nimpa.array2nii(
        resd, affined, Cntd['out']['fresdmu'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    # > joining bits together for rods in water
    resoW = np.concatenate((renW, resW, renW, buff), axis=0)
    # > padding for even dims
    # resoW = np.pad( resoW, ((1,0), (0,1),(0,1)), constant_values=((0,0), (0,0), (0, 0)) )
    resoW = np.pad(resoW,
                   ((1, 0), (Cntd['dpad'], Cntd['dpad'] + 1), (Cntd['dpad'], Cntd['dpad'] + 1)),
                   constant_values=((0, 0), (0, 0), (0, 0)))

    # > scale down
    resdW = ndi.zoom(resoW, 1 / Cntd['scld'], output=None, order=1, mode='constant')

    nimpa.array2nii(
        resdW, affined, Cntd['out']['fresdWmu'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    # > FOR registration using QNT reconstruction

    # > the extra part of the background at the bottom of the resolution bit after truncation.
    renB = renW[:bsz, ...].copy()
    renB[renB > 0] = 100

    renW[renW > Cntd['mu_water']] = 0
    renW[renW == Cntd['mu_water']] = 100

    resW[resW > Cntd['mu_water']] = 0
    resW[resW == Cntd['mu_water']] = 100

    # > joining bits together for rods in water
    resoW = np.concatenate((renW, resW, renW, renB), axis=0)
    # resoW = np.concatenate((renW, resW, renW), axis=0)

    # > padding for even dims
    resoW = np.pad(resoW,
                   ((1, 0), (Cntd['dpad'], Cntd['dpad'] + 1), (Cntd['dpad'], Cntd['dpad'] + 1)),
                   constant_values=((0, 0), (0, 0), (0, 0)))

    # > scale down
    resdW = ndi.zoom(resoW, 1 / Cntd['scld'], output=None, order=1, mode='constant')

    nimpa.array2nii(
        resdW, affined, Cntd['out']['fresdQmu'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    if return_raw:
        return {'resoW': resoW, 'resdW': resdW, 'resd': resd, 'reso': reso}


# =======================================================================
# >  S A M P L I N G   T E M P L A T E S
# =======================================================================


def create_sampl_reso(Cntd, return_raw=False):
    '''
    Create template of sampling rings for the resolution rods.
    Arguments:

    Cntd  - dictionary of constants, e.g.:
            spth  - output path of the folder used for storing the sampling
            templates
            dpth  - input path of the folder with the design in PNG format
            vxsz  - target voxel size
            scl   - scale factor from the original PNG voxel size to the target
            intord- interpolation order (0-nearest neighbour, etc.).
            rods_rotate - the initial rotation needed for accurately registering
                    the rods as they were placed during the phantom scan.
    imupd - dictionary of the trimmed and upsampled PET image used
            as references.
    '''

    if os.path.isfile(Cntd['out']['fst_res']):
        return None

    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        imupd = nimpa.getnii(Cntd['fqntup'], output='all')
    else:
        raise ValueError('Upscaled and trimmed ACR QNT PET image cannot be found')

    rods_png = imageio.imread(Cntd['fs_rods'])

    gamma = 1.
    y = np.int16(0.2126 * rods_png[..., 0]**gamma + 0.7152 * rods_png[..., 1]**gamma +
                 0.0722 * rods_png[..., 2]**gamma)

    # > unique values 11,10,9,7,6,4
    uy = np.unique(y)

    # > levels of rings for each rod diameter
    o = 6
    k = o
    lvls = []
    lvls.append(uy[:k])
    lvls.append(uy[k:k + o])
    k += o
    lvls.append(uy[k:k + o])
    k += o
    lvls.append(uy[k:k + o])
    k += o
    o = 5
    lvls.append(uy[k:k + o])
    k += o
    lvls.append(uy[k:k + o])
    k += o
    o = 4
    lvls.append(uy[k:k + o])
    k += o
    o = 3
    lvls.append(uy[k:k + o])
    k += o
    lvls.append(uy[k:k + o])
    k += o
    o = 2
    lvls.append(uy[k:k + o])
    k += o
    o = 1
    lvls.append(uy[k:k + o])
    k += o
    lvls.append(uy[k:k + o])
    k += o

    g = -1 * np.ones((6, len(lvls) - 1), dtype=np.int16)
    for i in range(len(lvls) - 1):
        lvls[i]
        # > go through 6 rod diameters
        g[:len(lvls[i]), i] = lvls[i]

    g = g[::-1, :]

    srods = np.zeros(rods_png[..., 0].shape, dtype=np.uint16)

    for i, off in enumerate([10, 20, 30, 40, 50, 60]):
        for j, lvl in enumerate(g[i, g[i, :] > 0]):
            srods[y == lvl] = off + j

    if Cntd['rods_rotate'] != 0:
        srods = ndi.rotate(srods, Cntd['rods_rotate'], reshape=False, order=Cntd['intord'],
                           mode='constant')

    srods = ndi.zoom(srods, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')

    # > rod ends
    k = int(np.round(Cntd['k_rodsend'] / Cntd['vxsz']))
    srend = np.zeros((k,) + srods.shape, dtype=np.uint8)

    # > rods themselves
    k = int(np.round(Cntd['k_rods'] / Cntd['vxsz']))
    srods = np.repeat(srods[None, ...], k, axis=0)

    sres = np.concatenate((srend, srods, srend), axis=0)
    sres = np.pad(sres, ((1, 0), (0, 1), (0, 1)), constant_values=((0, 0), (0, 0), (0, 0)))

    imxys = sres.shape[2]
    imzys = sres.shape[0]
    affine = np.array([[-Cntd['vxsz'], 0., 0., .5 * imxys * Cntd['vxsz']],
                       [0., Cntd['vxsz'], 0., -.5 * imxys * Cntd['vxsz']],
                       [0., 0., Cntd['vxsz'], -.5 * imzys * Cntd['vxsz']], [0., 0., 0., 1.]])

    nimpa.array2nii(
        sres, affine, Cntd['out']['fst_res'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    if return_raw:
        del srods, srend
        return sres
    else:
        del srods, sres, srend
        return None


def create_sampl(Cntd, return_raw=False):
    """Create template of sampling rings for the inserts."""
    if all(os.path.isfile(Cntd['out'][f] for f in ('fst_insrt', 'fst_insrt3', 'fst_ibckg'))):
        if return_raw:
            return {
                'allsmplng': Cntd['out']['fst_insrt'], 'insrt3smplng': Cntd['out']['fst_insrt3'],
                'ibckgsmplng': Cntd['out']['fst_ibckg']}
        return None

    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        imupd = nimpa.getnii(Cntd['fqntup'], output='all')
    else:
        raise ValueError('Upscaled and trimmed ACR QNT PET image cannot be found')

    # Unused for now:
    # bckg_png = imageio.imread(Cntd['fs_bckg'])
    # ibckg_png = imageio.imread(Cntd['fs_ibckg'])
    # air_png = imageio.imread(Cntd['fs_air'])
    # h2o_png = imageio.imread(Cntd['fs_h2o'])
    # bone_png = imageio.imread(Cntd['fs_bone'])
    # hot1_png = imageio.imread(Cntd['fs_hot1'])
    # hot2_png = imageio.imread(Cntd['fs_hot2'])
    # hot3_png = imageio.imread(Cntd['fs_hot3'])
    # hot4_png = imageio.imread(Cntd['fs_hot4'])

    # > converted and combined in fewer templates
    insrts = None

    tmpl = (k for k in Cntd if k[:3] == 'fs_')
    for k in tmpl:
        off = Cntd['soff_' + k[3:]]
        if off is None: continue
        print('file: {} >> offset = {}'.format(k, off))

        impng = imageio.imread(Cntd[k])

        # > array for conversion of labels
        arr = np.zeros(impng[..., 0].shape, dtype=np.uint16)

        for i, v in enumerate(np.unique(impng[..., 0])):
            if v < 255:
                arr += np.uint16((impng[..., 0] == v) * (off+i))

        # > assemble all the templates together apart from hot3 which is
        # > overlapping, hence separate
        if k[3:] == 'hot3':
            insrt3 = arr
        elif k[3:] == 'bckg':
            bckg = arr
        elif k[3:] == 'ibckg':
            ibckg = arr
        else:
            if insrts is None:
                insrts = arr
            else:
                insrts += arr

    # > AXIAL SPACING
    # > upper padding
    ku = int(
        sum(
            np.round(Cntd[k] / Cntd['vxsz'])
            for k in ('k_mcapA', 'k_mcapB', 'k_mcapC', 'k_mcapD', 'k_lidA', 'k_lidB')))
    ki = int(np.round(Cntd['k_insrtA'] / Cntd['vxsz']))

    # > insert bottoms
    kb = int(np.round(Cntd['k_insrtB'] / Cntd['vxsz']))

    # > uniform bit region
    kuni = int(np.round(Cntd['k_unfrm'] / Cntd['vxsz']))

    # > resolution region
    kr = int(np.round((Cntd['k_rods'] + 2 * Cntd['k_rodsend']) / Cntd['vxsz']))

    # > bottom fixed lid
    kl = int(np.round(Cntd['k_contB'] / Cntd['vxsz']))

    # > scale down the original PNG images
    insrts = ndi.zoom(insrts, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    insrt3 = ndi.zoom(insrt3, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    ibckg = ndi.zoom(ibckg, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')
    bckg = ndi.zoom(bckg, Cntd['scl'], output=None, order=Cntd['intord'], mode='constant')

    # > padding with zeros
    up_ = np.zeros((ku,) + insrts.shape, dtype=np.uint16)
    ibtm_ = np.zeros((kb,) + insrts.shape, dtype=np.uint16)
    btm_ = np.zeros((kl + kr,) + insrts.shape, dtype=np.uint16)

    # > repeat according to the axial dimensions
    insrts = np.repeat(insrts[None, ...], ki, axis=0)
    insrt3 = np.repeat(insrt3[None, ...], ki, axis=0)
    ibckg = np.repeat(ibckg[None, ...], ki, axis=0)
    bckg = np.repeat(bckg[None, ...], kuni, axis=0)

    allsmplng = np.concatenate((up_, insrts, ibtm_, bckg, btm_), axis=0)

    insrt3smplng = np.concatenate((up_, insrt3, ibtm_, bckg, btm_), axis=0)

    ibckgsmplng = np.concatenate((up_, ibckg, ibtm_, bckg, btm_), axis=0)

    # > pad the image to make the dims even
    allsmplng = np.pad(allsmplng, ((0, 0), (0, 1), (0, 1)),
                       constant_values=((0, 0), (0, 0), (0, 0)))
    insrt3smplng = np.pad(insrt3smplng, ((0, 0), (0, 1), (0, 1)),
                          constant_values=((0, 0), (0, 0), (0, 0)))
    ibckgsmplng = np.pad(ibckgsmplng, ((0, 0), (0, 1), (0, 1)),
                         constant_values=((0, 0), (0, 0), (0, 0)))

    # > make it up-side down (as it is scanned)
    allsmplng = allsmplng[::-1, :, :]
    insrt3smplng = insrt3smplng[::-1, :, :]
    ibckgsmplng = ibckgsmplng[::-1, :, :]

    # > get the affine and save to NIfTI
    imxys = ibckgsmplng.shape[2]
    imzys = ibckgsmplng.shape[0]
    affine = np.array([[-Cntd['vxsz'], 0., 0., .5 * imxys * Cntd['vxsz']],
                       [0., Cntd['vxsz'], 0., -.5 * imxys * Cntd['vxsz']],
                       [0., 0., Cntd['vxsz'], -.5 * imzys * Cntd['vxsz']], [0., 0., 0., 1.]])

    nimpa.array2nii(
        allsmplng, affine, Cntd['out']['fst_insrt'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    nimpa.array2nii(
        insrt3smplng, affine, Cntd['out']['fst_insrt3'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    nimpa.array2nii(
        ibckgsmplng, affine, Cntd['out']['fst_ibckg'],
        trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
               imupd['transpose'].index(2)), flip=imupd['flip'])

    if return_raw:
        return {'allsmplng': allsmplng, 'insrt3smplng': insrt3smplng, 'ibckgsmplng': ibckgsmplng}



"""
Functions to help imaging with the PET/MR GE Signa scanner
"""

__author__ = "Pawel Markiewicz"
__copyright__ = "Copyright 2023"

import shutil
from pathlib import Path

import h5py
import nibabel as nib
import numpy as np

from ..prc import imio

SZ_VXZ = 2.78      # Signa default axial voxel size


def pifa2nii(fpifa, fnii=None, outpath=None):
    ''' Convert the GE PIFA file format to NIfTI. For generating PIFA
        file in the same space as the PET reconstructed image, a NIfTI
        file of the reconstruction needs to be provided as `fnii`.

        The NIfTI file needs to represent the whole FOV of 60 cm.
    '''

    fpifa = Path(fpifa)
    if not fpifa.is_file():
        raise FileNotFoundError('PIFA file could not be found')

    if fnii is not None and Path(fnii).is_file():
        affine = imio.getnii(fnii, output='affine')
    else:
        affine = None

    # > output folder
    if outpath is None:
        pifadir = fpifa.parent
    else:
        pifadir = Path(outpath)
    imio.create_dir(pifadir)

    # > read the HDF5 file
    fh = h5py.File(fpifa, 'r')

    # > Diameter of the transaxial FOV
    DFOV = fh['HeaderData/ctacDfov'][0]

    # > PIFA x-y voxel size
    SZ_IMX = fh['HeaderData/xMatrix'][0]
    SZ_IMZ = fh['HeaderData/zMatrix'][0]
    # ZLOCAT = fh['HeaderData/tableLocation'][0] # << check if this is really correct
    SP_VXY = DFOV / SZ_IMX

    # > the mu-map in units of 1/mm
    mu_dat = np.array(fh['PifaData'])

    # > transpose the data for NIfTI output
    data = np.transpose(mu_dat[:, ::-1, :], (2, 1, 0))

    # > if affine is not provided through the NIfTI
    if affine is None:
        # > affine matrix
        A = np.eye(4)
        A[0, 0] = -SP_VXY
        A[1, 1] = SP_VXY
        A[2, 2] = SZ_VXZ

        A[0, 3] = DFOV/2 - SP_VXY/2
        A[1, 3] = -(DFOV/2 - SP_VXY/2)
        A[2, 3] = -(SZ_IMZ/2*SZ_VXZ - SZ_VXZ/2)
        A[3, 3] = 1
    else:
        # A = np.eye(4)
        # A[0,0] = -SP_VXY
        # A[1,1] = SP_VXY
        # A[2,2] = SZ_VXZ

        # A[0,3] = affine[0,3]
        # A[1,3] = affine[1,3]
        # A[2,3] = affine[2,3]
        # A[3,3] = 1
        A = affine

    img = nib.Nifti1Image(data, A)
    fout = pifadir / (fpifa.name.split('.pifa')[0] + '.nii.gz')
    img.to_filename(fout)

    return fout


def nii2pifa(fnii, fpifa, outpath=None, bed_mask_thresh=0.2):
    '''
    Convert a newly generated PIFA NIfTI `fnii` file to the
    GE PIFA file format using the original PIFA file `fpifa`.
    '''

    fpifa = Path(fpifa)
    if 'pifaIvv_' in fpifa.name:
        fpifa_ivv = fpifa
        fpifa = fpifa.parent / fpifa.name.replace('pifaIvv_', 'pifa_')
    elif 'pifa_' in fpifa.name:
        fpifa_ivv = fpifa.parent / fpifa.name.replace('pifa_', 'pifaIvv_')
    else:
        raise FileNotFoundError('PIFA files are unrecognised')

    # > output folder
    pifadir = fpifa.parent if outpath is None else Path(outpath)
    imio.create_dir(pifadir)

    # > new PIFA
    fpifa_n = pifadir / fpifa.name
    fpifa_ivv_n = pifadir / fpifa_ivv.name
    shutil.copyfile(fpifa, fpifa_n)
    shutil.copyfile(fpifa_ivv, fpifa_ivv_n)

    with h5py.File(fpifa_n, 'r+') as fh, h5py.File(fpifa_ivv_n, 'r+') as fh_ivv:
        # > get all the components of original PIFA
        pifa = np.array(fh['PifaData'])
        pifa_ivv = np.array(fh_ivv['PifaData'])
        bed = pifa - pifa_ivv
        msk_bed = bed > bed_mask_thresh * np.max(bed)

        # > make the new mu-map to PIFA
        newpifa_ivv = 0.1 * imio.getnii(fnii)
        # > flip the z-axis (in GE systems it's the other way round)
        # > and set to zero the voxels which belong to the bed/table
        newpifa_ivv = newpifa_ivv[::-1, ...] * ~msk_bed

        # > blend the bed and object
        newpifa = np.maximum(bed, newpifa_ivv)

        fh['PifaData'][...] = newpifa
        fh_ivv['PifaData'][...] = newpifa_ivv

    # > check if edited:
    with h5py.File(fpifa_n, 'r') as fcheck:
        dcheck = fcheck['PifaData'][...]
    if not np.allclose(dcheck, newpifa):
        raise ValueError('the CT modification did not work')

    with h5py.File(fpifa_ivv_n, 'r') as fcheck:
        dcheck = fcheck['PifaData'][...]
    if not np.allclose(dcheck, newpifa_ivv):
        raise ValueError('the CT modification did not work')

    return {'fpifa': fpifa_n, 'fpifa_ivv': fpifa_ivv_n, 'pifa': newpifa, 'opifa': pifa}



# initialise the module folder
__all__ = [
    # imio
    'array2nii', 'create_dir', 'dcm2im', 'dcm2nii', 'dcmanonym', 'dcminfo', 'dcmsort', 'fwhm2sig',
    'mgh2nii', 'getmgh', 'getnii', 'getnii_descr', 'nii_gzip', 'nii_ugzip', 'niisort', 'orientnii',
    'pick_t1w', 'time_stamp', 'rem_chars', 'isdcm', 'dcmdir',
    # prc
    'bias_field_correction', 'centre_mass_img', 'centre_mass_rel', 'centre_mass_corr', 'ct2mu',
    'im_cut', 'imsmooth', 'imtrimup',
    'iyang', 'nii_modify', 'pet2pet_rigid', 'psf_gaussian', 'psf_measured', 'pvc_iyang',
    # num
    'conv_separable', 'isub', 'nlm',
    # regseg
    'aff_dist', 'affine_dipy', 'affine_fsl', 'affine_niftyreg',
    'coreg_spm', 'coreg_vinci', 'create_mask', 'dice_coeff', 'dice_coeff_multiclass',
    'imfill', 'motion_reg', 'realign_mltp_spm', 'resample_fsl', 'resample_dipy',
    'resample_mltp_spm', 'resample_niftyreg', 'resample_spm',
    'resample_vinci'] # yapf: disable

from .imio import (
    array2nii,
    create_dir,
    dcm2im,
    dcm2nii,
    dcmanonym,
    dcmdir,
    dcminfo,
    dcmsort,
    fwhm2sig,
    getmgh,
    getnii,
    getnii_descr,
    isdcm,
    mgh2nii,
    nii_gzip,
    nii_ugzip,
    niisort,
    orientnii,
    pick_t1w,
    rem_chars,
    time_stamp,
)
from .num import conv_separable, isub, nlm

# will be deprecated
from .prc import (
    bias_field_correction,
    centre_mass_corr,
    centre_mass_img,
    centre_mass_rel,
    ct2mu,
    im_cut,
    imsmooth,
    imtrimup,
    iyang,
    nii_modify,
    pet2pet_rigid,
    psf_gaussian,
    psf_measured,
    pvc_iyang,
)
from .regseg import (
    aff_dist,
    affine_dipy,
    affine_fsl,
    affine_niftyreg,
    coreg_spm,
    coreg_vinci,
    create_mask,
    dice_coeff,
    dice_coeff_multiclass,
    imfill,
    motion_reg,
    realign_mltp_spm,
    resample_dipy,
    resample_fsl,
    resample_mltp_spm,
    resample_niftyreg,
    resample_spm,
    resample_vinci,
)



# initialise the module folder
__all__ = [
    'get_params', 'get_paths', 'extract_reso_part', 'sampling_masks', 'create_mumap_core',
    'create_nac_core', 'create_reso', 'create_sampl_reso', 'create_sampl', 'standard_analysis',
    'estimate_fwhm', 'preproc']

from .analysis import estimate_fwhm, standard_analysis
from .ioaux import extract_reso_part, get_paths, sampling_masks
from .params import get_params
from .proc import preproc
from .templates import (
    create_mumap_core,
    create_nac_core,
    create_reso,
    create_sampl,
    create_sampl_reso,
)



"""ACR/Jaszczak PET phantom design parameters"""
__author__ = "Pawel Markiewicz"
__copyright__ = "Copyright 2021-23"

import os
from pathlib import Path

import numpy as np

try:          # py<3.9
    import importlib_resources as resources
except ImportError:
    from importlib import resources

# CONSTANTS/PARAMTERS

# smoothing of the templates before registration
fwhm_tmpl = 3.

# smoothing kernel FWHM for the NAC and QNT PET
fwhm_nac = 3.
fwhm_qnt = 3.

# number of iterations for NAC and QNT PET reconstruction
itr_nac = 1
itr_qnt = 4
itr_qnt2 = [4, 8, 16]

# image trimming upsampling scaling factor
sclt = 4

# and interpolation (1: linear)
interp = 1

# transaxial voxel size of the PNG images: 300dpi => 0.08467 mm
vxysz = 25.4 / 300
# target voxel size
vxsz = 0.2

# scaling factor for images from PNG image resolution
scl = vxysz / vxsz

# second level image scaling factor
scld = 2

# interpolation order (0-NN, 1-trilinear)
intord = 0

# padding for the design arrays (gives more breathing space for smoothing in transaxial views)
pad = 4

# angle of rotations for the initial position of resolution rods (cold)
rods_rotate = 0

# > buffer size for the activity image for registration purposes
buff_size = 16

# mu-value for the perspex attenuation
mu_prspx = 0.1036
mu_water = 0.096
mu_bone = 0.148
mu_screw = 0.117

# PNG values attenuation
water = 209
prspx = 167
scrws = 129
boney = 113

# activity values for inserts and background and NAC edges
ains = 200
abck = 100
aedg = 150

# PNG values for activity
png_ains = 129
png_abck = 167
png_acap = 129
png_aedg = 167

# ACR ring sample distancing
# background circle radii:
rbckgs = np.arange(12, 220, 12) / 2
# sampling ticks
sbckgs = rbckgs - rbckgs[0] / 2

# insert sampling
rinsrt = np.array([4, 8, 12, 16, 20, 25, 30, 34, 38, 42, 46, 50]) / 2
sinsrt = rinsrt - rinsrt[0] / 2
sinsrt[0] = 0.

# sampling resolution rods
# centres of sampling rings
srngs = np.array([2.52, 4.53, 6.54, 8.56, 10.65, 12.68, 14.71, 16.74, 18.8, 20.8, 24]) / 2
rods_rngc = np.concatenate(([0], srngs[0:-1] + (srngs[1:] - srngs[0:-1]) / 2))

# resolution rod radii
rods_rad = np.array([12.6, 11.02, 9.46, 7.84, 6.27, 4.66]) / 2

# rods nominal diameters
rods_nom = np.array(['12.7', '11.1', '9.5', '7.9', '6.4', '4.8'])

# number of sampling rings for each rod
rods_nrngs = np.array([11, 10, 9, 7, 6, 4])

# sampling offset for each rod pie-piece
rods_off = np.array([60, 50, 40, 30, 20, 10])

# DIPY rigid body registration parameters
dipy_itrs = [10000, 1000, 200]
dipy_sgms = [3.0, 1.0, 0.0]
dipy_fcts = [4, 2, 1]

# for rods more iterations
dipy_rods_itrs = [10000, 1000, 300]


def get_params(cpath=None):
    """
    Get all the parameters and file names for ACR template generation.
    Arguments:
      cpath  : path of custom ACR design files
    """
    if cpath is None:
        f_core_main_comp = resources.files(
            "niftypet.nimpa").resolve() / "acr_design" / "core_mumap" / "acr-main-compartment.png"
        if not f_core_main_comp.is_file():
            raise IOError('Design ACR phantom files are missing in the NIMPA package')
        mfldr = f_core_main_comp.parent.parent

    elif os.path.isdir(cpath) and os.path.isfile(
            os.path.join(cpath, 'core_mumap', 'acr-main-compartment.png')):
        mfldr = Path(cpath)
    else:
        raise ValueError('The input folder is not recognised')

    # design PDF/PNG path for sampling, mu-map, NAC distribution
    dmpth = mfldr / 'core_mumap'
    dnpth = mfldr / 'core_nac'
    drpth = mfldr / 'rods'
    dspth = mfldr / 'sampling'

    # dictionary of design constant
    Cntd = {
        'dspth': dspth, 'dmpth': dmpth, 'vxysz': vxysz, 'vxsz': vxsz, 'sclt': sclt,
        'interp': interp, 'scld': scld, 'scl': scl, 'intord': intord, 'rods_rotate': rods_rotate,
        'buff_rods_size': buff_size, 'itr_nac': itr_nac, 'itr_qnt': itr_qnt, 'itr_qnt2': itr_qnt2,
        'fwhm_nac': fwhm_nac, 'fwhm_qnt': fwhm_qnt}
    # mu-values
    Cntd.update(mu_prspx=mu_prspx, mu_water=mu_water, mu_bone=mu_bone, mu_screw=mu_screw)

    # activity values
    Cntd.update(ains=ains, abck=aedg)
    # PNG mu-values
    Cntd.update(png_water=water, png_prspx=prspx, png_scrwy=boney)
    # PNG activity values
    Cntd.update(png_ains=png_ains, png_abck=png_abck, png_acap=png_acap, png_aedg=png_aedg,
                dpad=pad, fwhm_tmpl=fwhm_tmpl, dipy_itrs=dipy_itrs, dipy_sgms=dipy_sgms,
                dipy_fcts_itrs=dipy_rods_itrs)
    # ACR core mu-map designs
    Cntd.update(fmcap0=dmpth / 'acr-main-cap-0.png', fmcap1=dmpth / 'acr-main-cap-1.png',
                fmcap2=dmpth / 'acr-main-cap-2.png', fbscrw=dmpth / 'acr-bone-screw.png',
                fscrws=dmpth / 'acr-screws.png', flid0=dmpth / 'acr-lid0.png',
                flid1=dmpth / 'acr-lid1.png', finsrt=dmpth / 'acr-inserts.png',
                fbinsrt=dmpth / 'acr-inserts-bottoms.png', fmain=dmpth / 'acr-bottom.png')
    # ACR core activity (NAC) designs
    Cntd.update(fcap=dnpth / 'acr-cap.png', fins=dnpth / 'acr-inserts.png',
                fbig=dnpth / 'acr-rng.png')
    # ACR rods
    Cntd.update(frespng=drpth / 'acr-rods.png', frenpng=drpth / 'acr-rods-ends.png',
                fresWpng=drpth / 'acr-rods-ends-water.png')
    # ACR sampling designs
    Cntd.update(fs_rods=dspth / 'acr-rods-sampling.png', fs_bckg=dspth / 'acr-bckg-sampling.png',
                fs_ibckg=dspth / 'acr-insrt-bckg-sampling.png',
                fs_air=dspth / 'acr-air-sampling.png', fs_h2o=dspth / 'acr-h2o-sampling.png',
                fs_bone=dspth / 'acr-bone-sampling.png', fs_hot1=dspth / 'acr-hot1-sampling.png',
                fs_hot2=dspth / 'acr-hot2-sampling.png', fs_hot3=dspth / 'acr-hot4-sampling.png')
    # parameters for plotting rods results
    Cntd.update(rods_rngc=rods_rngc, rods_rad=rods_rad, rods_nom=rods_nom, rods_nrn=rods_off)
    # intensity offsetsfor the front phantom sampling
    Cntd.update(soff_rods=None, soff_bckg=300, soff_ibckg=200, soff_air=70, soff_h2o=50,
                soff_bone=90, soff_hot1=10, soff_hot2=20, soff_hot3=30, soff_hot4=40,
                rbckgs=rbckgs, sbckgs=sbckgs, rinsrt=sinsrt)
    # AXIAL DEMNSIONS, extension in mm
    Cntd.update(k_rods=83.40, k_rodsend=3.0) # rods

    # CORE PHANTOM
    Cntd.update(
        k_mcapA=19.2,   # main cap top
        k_mcapB=5.3,    # main cap
        k_mcapC=6.7,    # main cap
        k_mcapD=12.3,   # main cap bottom
        k_lidA=5.8,     # lid top
        k_lidB=5.8,     # lid bottom
        k_insrtA=38.0,  # inserts
        k_insrtB=1.5,   # insert bottoms
        k_contA=147.5,  # main container
        k_unfrm=58.101, # main container uniform part
        k_contB=11.6,   # container bottom
    )

    return Cntd



"""ACR/Jaszczak PET phantom I/O and auxiliary functions"""
__author__ = "Pawel Markiewicz"
__copyright__ = "Copyright 2021-23"
import os
from pathlib import Path, PurePath

import dipy.align as align
import nibabel as nib
import numpy as np

try:
    import matplotlib.patches as patches
    import matplotlib.pyplot as plt
except ImportError:
    pass

from ..prc import imio

# INPUT/OUTPUT


def get_paths(Cntd, outpath=None):
    '''
    Get all the paths for generating mu-maps and sampling templates.
    Arguments:
    Cntd    - Dictionary of constants and parameters
    outpath - output folder/path
    '''

    fimup = None
    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        fimup = Path(Cntd['fqntup'])

    elif 'fnacup' in Cntd and Path(Cntd['fnacup']).is_file():
        fimup = Path(Cntd['fnacup'])

    else:
        raise ValueError(
            'NAC or QNT PET image has to be identified in the constants dictionary Cntd')

    if not fimup:
        return None

    # > output path
    if isinstance(outpath, (str, PurePath)):
        opth = outpath
        Cntd['opth'] = opth
    elif 'opth' in Cntd:
        opth = Cntd['opth']
    else:
        raise ValueError('unrecognised output path')

    imio.create_dir(opth)

    # folder with templates, which will be registered to PET images
    tfldr = os.path.join(opth, 'templates')
    imio.create_dir(tfldr)
    t_acr_core_flddr = os.path.join(tfldr, 'ACR-core')
    imio.create_dir(t_acr_core_flddr)
    t_acr_reso_flddr = os.path.join(tfldr, 'ACR-rods')
    imio.create_dir(t_acr_reso_flddr)
    t_acr_smpl_flddr = os.path.join(tfldr, 'ACR-smpl')
    imio.create_dir(t_acr_smpl_flddr)

    # target rods output folder path
    rpth = os.path.join(opth, 'ACR-rods')
    imio.create_dir(rpth)
    # target core output folder path
    mpth = os.path.join(opth, 'ACR-core')
    imio.create_dir(mpth)

    # -----------------------------------------------------------
    # Core ACR and NAC PET (for registration)
    # -----------------------------------------------------------
    # rigid transformation affine output file name
    # for the NAC main/core part
    faff = os.path.join(mpth, 'affine-acr-dipy.npy')

    # output file name for the ACR NAC activity for registration
    facrad = os.path.join(
        t_acr_core_flddr,
        'acr-activity-' + str(Cntd['vxsz'] * Cntd['scld']).replace('.', '-') + 'mm.nii.gz')

    # output file for the ACR mu-map without resolution rod insert
    facrmu = os.path.join(t_acr_core_flddr,
                          'acr-umap-' + str(Cntd['vxsz']).replace('.', '-') + 'mm.nii.gz')

    # the same as above but with lower resolution (double voxel size)
    facrdmu = os.path.join(
        t_acr_core_flddr,
        'acr-umap-' + str(Cntd['vxsz'] * Cntd['scld']).replace('.', '-') + 'mm.nii.gz')

    # output mu-map for the original PET image reconstruction
    fmuo = os.path.join(mpth, 'acr-mumap-dipy.nii.gz')

    # output mu-map for the original PET image reconstruction
    # (high resolution, i.e., smaller voxel size)
    fmuo_hires = os.path.join(mpth, 'acr-mumap-hires-dipy.nii.gz')

    # full mu-map (combined core and resolution parts) - the endpoint result
    fmuf = os.path.join(opth, 'acr-complete-umap.nii.gz')

    # -----------------------------------------------------------
    # RESOLUTION RODS
    # -----------------------------------------------------------
    # output file name for the resolution rods mu-map
    fresomu = os.path.join(t_acr_reso_flddr,
                           'acr-reso-umap-' + str(Cntd['vxsz']).replace('.', '-') + 'mm.nii.gz')

    # the same but double voxel size (as defined by scld)
    fresdmu = os.path.join(
        t_acr_reso_flddr,
        'acr-reso-umap-' + str(Cntd['vxsz'] * Cntd['scld']).replace('.', '-') + 'mm.nii.gz')

    # the resolution rods mu-map with water background (used for the phantom)
    fresdWmu = os.path.join(
        t_acr_reso_flddr,
        'acr-reso-water-umap-' + str(Cntd['vxsz'] * Cntd['scld']).replace('.', '-') + 'mm.nii.gz')

    # the resolution active QNT for alternative registration of the rods to PET
    fresdQmu = os.path.join(
        t_acr_reso_flddr,
        'acr-reso-active-qnt-' + str(Cntd['vxsz'] * Cntd['scld']).replace('.', '-') + 'mm.nii.gz')

    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        # output file for the resolution rods part only
        fpet_res = os.path.join(
            rpth,
            os.path.basename(fimup).split('-scale-' + str(Cntd['sclt']) + '-')[0] + '_rods.nii.gz')

        # output file for rigid body transformation affine
        faff_res = os.path.join(rpth, 'affine-dipy-acr-reso.npy')
        # if os.path.isfile(faff_res):
        #     txaff = np.load(faff_res)

        fmur = os.path.join(rpth, 'acr-reso-water-mumap-dipy.nii.gz')
    else:
        fpet_res = ''
        faff_res = ''
        fmur = ''

    # -----------------------------------------------------------
    # SAMPLING
    # -----------------------------------------------------------
    # sampling ring template for the rods
    fst_res = os.path.join(t_acr_smpl_flddr,
                           'acr-res-sampling-' + str(Cntd['vxsz']).replace('.', '-') + 'mm.nii.gz')
    # sampling template for the inserts
    fst_insrt = os.path.join(
        t_acr_smpl_flddr, 'acr-all-sampling-' + str(Cntd['vxsz']).replace('.', '-') + 'mm.nii.gz')
    # sampling template for the 3rd hot insert
    fst_insrt3 = os.path.join(
        t_acr_smpl_flddr,
        'acr-insrt3-sampling-' + str(Cntd['vxsz']).replace('.', '-') + 'mm.nii.gz')
    # sampling template for the insert background
    fst_ibckg = os.path.join(
        t_acr_smpl_flddr,
        'acr-ibckg-sampling-' + str(Cntd['vxsz']).replace('.', '-') + 'mm.nii.gz')

    # get the output file names into the dictionary
    Cntd['out'] = {
        'faff_res': faff_res, 'faff': faff, 'fpet_res': fpet_res, 'fresomu': fresomu,
        'fresdmu': fresdmu, 'fresdWmu': fresdWmu, 'fresdQmu': fresdQmu, 'fmur': fmur,
        'facrad': facrad, 'facrmu': facrmu, 'facrdmu': facrdmu, 'fmuo': fmuo,
        'fmuo_hires': fmuo_hires, 'fmuf': fmuf, 'fst_res': fst_res, 'fst_insrt': fst_insrt,
        'fst_insrt3': fst_insrt3, 'fst_ibckg': fst_ibckg}

    return Cntd


def extract_reso_part(Cntd, offset=15, forced=False):
    '''
    extract resolution part for registration
    Arguments:
    offset:     the extend of the extra extension from rods to the uniform
                region.
    forced:     force the extraction even if detected to be already done.
    '''

    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        imupd = imio.getnii(Cntd['fqntup'], output='all')
    else:
        raise ValueError('Upscaled and trimmed QNT ACR PET image cannot be found')

    if forced or not os.path.isfile(Cntd['out']['fpet_res']):
        # pick the axial cut-off position based on the axial summed profile
        axprf = np.sum(imupd['im'], axis=(1, 2))

        thrshld = np.max(axprf) * 0.33
        ids = np.where(axprf > thrshld)
        i0 = np.min(ids)
        i1 = axprf.argmax()

        tmp = axprf[i0:i1]
        wndw = 7
        dip = np.array([(tmp[i + wndw] - tmp[i]) < -4e6 for i in range(i1 - i0 - wndw)])

        cutoff = np.max(np.where(dip)) + i0 + offset
        '''
        figure(); plot(axprf, '.-')
        figure(); plot(tmp, '.-')
        '''

        rodsim = imupd['im'].copy()
        rodsim[cutoff:, ...] = 0

        imio.array2nii(
            rodsim, imupd['affine'], Cntd['out']['fpet_res'],
            trnsp=(imupd['transpose'].index(0), imupd['transpose'].index(1),
                   imupd['transpose'].index(2)), flip=imupd['flip'])

        return tmp

    else:
        print('i> extraction has already been performed')
        return None


def sampling_masks(Cntd, use_stored=False):
    ''' get the sampling masks for analysis of the ACR phantom
    '''

    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        fimup = str(Cntd['fqntup'])
    else:
        raise ValueError('Upscaled and trimmed ACR PET image cannot be found')

    # prepare output folder
    smpl_dir = os.path.join(os.path.dirname(Cntd['out']['fmuf']), 'sampling_masks')
    imio.create_dir(smpl_dir)
    print(f'i> using sampling folder: {smpl_dir}')

    # > dictionary of VOIs for sampling PET images,
    # > made by registering the templates to PET space.
    vois = {}

    for t in (t for t in Cntd['out'] if t[:4] == 'fst_'):
        fpth = Cntd['out'][t]

        print(fpth)

        if not os.path.isfile(fpth):
            raise ValueError('The sampling template file does not exists.')

        # template in PET space
        fvois = os.path.join(smpl_dir, os.path.basename(fpth).split('.nii.gz')[0] + '_dipy.nii.gz')

        # if the resampled template does not exist, resample it
        if not use_stored or not os.path.isfile(fvois):
            if t == 'fst_res':
                faff = Cntd['out']['faff_res']
            else:
                faff = Cntd['out']['faff']

            print('i> using this affine for resampling:\n   ', faff)

            affine = np.load(faff)

            static, static_affine, moving, moving_affine, between_affine = (
                align._public._handle_pipeline_inputs(fpth, fimup, moving_affine=None,
                                                      static_affine=None, starting_affine=affine))

            affine_map = align._public.AffineMap(between_affine, static.shape, static_affine,
                                                 moving.shape, moving_affine)

            rsmpl = affine_map.transform(moving, interpolation='nearest')

            smpl_nii = nib.Nifti1Image(np.int32(rsmpl), static_affine)
            nib.save(smpl_nii, fvois)

            del moving, static

        vois[t] = imio.getnii(fvois)

        # > masks/vois for standard analysis of the ACR phantom
        # > also get the axial range of each concentric VOIs

        # > 6 cm diameter ROI for ACR background
        if t == 'fst_ibckg':
            # > get the range for the uniform part
            zax = np.sum(vois[t] == 317, axis=(1, 2))
            zax = np.where(zax > 0)[0]
            vois['r_bckg'] = [zax[0], zax[-1]]

            # > get the range for the faceplate inserts
            zax = np.sum(vois[t] == 217, axis=(1, 2))
            zax = np.where(zax > 0)[0]
            vois['r_ibckg'] = [zax[0], zax[-1]]

            # > get the standard ACR background VOI
            vois['s_bckg'] = (vois[t] >= 200) & (vois[t] < 205)

        elif t == 'fst_insrt':
            # > get the range:
            zax = np.sum(vois[t] == 100, axis=(1, 2))
            zax = np.where(zax > 0)[0]
            vois['r_insrt'] = [zax[0], zax[-1]]

            # > get the VOI for the inserts
            vois['s_i1'] = (vois[t] <= 14) & (vois[t] >= 10)
            vois['s_i2'] = (vois[t] <= 24) & (vois[t] >= 20)
            vois['s_i4'] = (vois[t] <= 44) & (vois[t] >= 40)
            vois['s_w'] = (vois[t] <= 54) & (vois[t] >= 50)
            vois['s_a'] = (vois[t] <= 74) & (vois[t] >= 70)
            vois['s_b'] = (vois[t] <= 94) & (vois[t] >= 90)

        elif t == 'fst_insrt3':
            # > get the range:
            zax = np.sum(vois[t] == 38, axis=(1, 2))
            zax = np.where(zax > 0)[0]
            vois['r_insrt3'] = [zax[0], zax[-1]]

            # > get the VOI for the insert
            vois['s_i3'] = (vois[t] <= 34) & (vois[t] >= 30)

        elif t == 'fst_res':
            # > get the range:
            zax = np.sum(vois[t] == 70, axis=(1, 2))
            zax = np.where(zax > 0)[0]
    return vois


def zmask(masks, key, Cntd, axial_offset=8, width_mm=10, z_start_idx=None, level=None):
    '''
    Obtain a sub-mask for the standard ACR analysis using 1 cm slice.
    Arguments:
        width_mm: axial width of the ROI mask (in mm)
    '''

    if 'fqntup' in Cntd and Path(Cntd['fqntup']).is_file():
        imupd = imio.getnii(Cntd['fqntup'], output='all')
    else:
        raise ValueError('Upscaled and trimmed ACR PET image cannot be found')

    if level is None:
        zax = np.sum(masks[key], axis=(1, 2))
    else:
        zax = np.sum(masks[key] == level, axis=(1, 2))
    zax = np.where(zax > 0)[0]

    # width of axial voxel extension
    width_vox = int(np.round(width_mm / imupd['voxsize'][0]))

    # the range
    z0 = zax[0] + axial_offset
    z1 = zax[-1] - axial_offset # wip: maybe -width_vox

    msk = np.zeros(masks[key].shape, dtype=bool)

    if z_start_idx is None:
        z_start_idx = z0

    msk[z_start_idx:z_start_idx + width_vox, ...] = True

    zmasks = {
        'z0': z0, 'z1': z1, 'zax': zax, 'width_vox': width_vox, 'z_start_idx': z_start_idx,
        'msk': msk}
    zmasks[key] = masks[key] * msk

    if key == 'fst_insrt':
        # 6 cm diameter ROI for ACR background
        msk_bck = (masks['fst_ibckg'] >= 200) & (masks['fst_ibckg'] < 205) * zmasks['msk']
        # all the inserts
        msk_i1 = (masks['fst_insrt'] <= 14) & (masks['fst_insrt'] >= 10) * zmasks['msk']
        msk_i2 = (masks['fst_insrt'] <= 24) & (masks['fst_insrt'] >= 20) * zmasks['msk']
        msk_i3 = (masks['fst_insrt3'] <= 34) & (masks['fst_insrt3'] >= 30) * zmasks['msk']
        msk_i4 = (masks['fst_insrt'] <= 44) & (masks['fst_insrt'] >= 40) * zmasks['msk']
        msk_w = (masks['fst_insrt'] <= 54) & (masks['fst_insrt'] >= 50) * zmasks['msk']
        msk_a = (masks['fst_insrt'] <= 74) & (masks['fst_insrt'] >= 70) * zmasks['msk']
        msk_b = (masks['fst_insrt'] <= 94) & (masks['fst_insrt'] >= 90) * zmasks['msk']

        zmasks['msk_bckg'] = msk_bck
        zmasks['msk_i1'] = msk_i1
        zmasks['msk_i2'] = msk_i2
        zmasks['msk_i3'] = msk_i3
        zmasks['msk_i4'] = msk_i4
        zmasks['msk_w'] = msk_w
        zmasks['msk_a'] = msk_a
        zmasks['msk_b'] = msk_b

        zmasks['fst_ibckg'] = masks['fst_ibckg'] * msk

        zmasks[key + '3'] = masks[key + '3'] * msk

    return zmasks


def extract_rings(img, tmpl, l0=None, l1=None):
    """
    extract average ROI ring values from PET image using the high resolution templates.
    """

    tmpl = np.int32(tmpl)

    # unique ring labels (id values)
    urs = np.unique(tmpl)

    if l0 is None:
        l0 = np.min(urs)
    if l1 is None:
        l1 = np.max(urs) + 1

    # rings ids (labels)
    irngs = range(l0, l1)

    # number of rings (labels)
    nrng = len(irngs)

    # initialise array for storing average voxel values for any label
    vrngs = np.zeros(nrng, dtype=np.float32)

    for i, l in enumerate(irngs):
        msk = tmpl == l
        vrngs[i] = np.sum(img[msk]) / np.sum(msk)

    return vrngs


def plot_hotins(im, masks, Cntd, inserts=None, axes=None, ylim=None, colour='k', marker_sz=3.5,
                draw_bckg=True, legend=True):
    """
    plot sampling rings in hot insert regions of size 25, 16, 12 and 8mm.
    Arguments:
      im:     input image for sampling (high resolution of around 0.5 mm)
      masks:  the concentric VOIs for sampling inserts
      inserts: the list of selected inserts for plotting
              (1: the biggest, 4: the smallest)
      draw_back: if True, draws insert walls in the background
    """
    if inserts is None:
        inserts = [1, 2, 3, 4]

    rinsrt = [12.5, 8, 6, 4]

    if axes is None:
        _, ax = plt.subplots(1)
    else:
        ax = axes

    hots = np.array([])
    # extract ring values
    if 1 in inserts:
        rh1 = extract_rings(im, masks['fst_insrt'], l0=10, l1=20)
        hots = np.concatenate((rh1, hots), axis=0)
        plt.plot(Cntd['sinsrt'][:-2], rh1, 'o-', color=colour, ms=marker_sz, label='25-mm')

    if 2 in inserts:
        rh2 = extract_rings(im, masks['fst_insrt'], l0=20, l1=30)
        hots = np.concatenate((rh2, hots), axis=0)
        plt.plot(Cntd['sinsrt'][:-2], rh2, 'p-', color=colour, ms=marker_sz, label='16-mm')

    if 3 in inserts:
        rh3 = extract_rings(im, masks['fst_insrt3'], l0=30, l1=40)
        hots = np.concatenate((rh3, hots), axis=0)
        plt.plot(Cntd['sinsrt'][:-2], rh3, 'x-', color=colour, ms=marker_sz, label='12-mm')

    if 4 in inserts:
        rh4 = extract_rings(im, masks['fst_insrt'], l0=40, l1=50)
        hots = np.concatenate((rh4, hots), axis=0)
        plt.plot(Cntd['sinsrt'][:-2], rh4, 'v-', color=colour, ms=marker_sz, label='8-mm')

    # get the global y-limits for all axes
    if ylim is None:
        ylim = [np.floor(np.amin(hots) / 100) * 100, np.ceil(np.amax(hots) / 100) * 100]

    if draw_bckg:
        for i in inserts:
            rect = patches.Rectangle((rinsrt[i - 1], ylim[0]), 1.5, ylim[1], linewidth=0,
                                     edgecolor='None', facecolor='0.85')
            ax.add_patch(rect)

    ax.set_ylim(ylim)
    plt.ylabel('Bq/mL')
    plt.xlabel('sampling ring radii [mm]')
    plt.title('hot inserts')
    if legend:
        plt.legend()
    return ax


def plot_coldins(im, Cntd, masks, ylim=None, line_style='.-', colour='k', axes=None,
                 draw_bckg=True):
    """plot sampling rings in cold insert regions of water, air and bone"""
    if axes is None:
        _, axs = plt.subplots(1, 3)
    else:
        axs = axes

    colds = np.array([])

    r_h2o = extract_rings(im, masks['fst_insrt'], l0=50, l1=62)
    r_air = extract_rings(im, masks['fst_insrt'], l0=70, l1=82)
    r_bone = extract_rings(im, masks['fst_insrt'], l0=90, l1=102)
    colds = np.concatenate((colds, r_h2o, r_air, r_bone), axis=0)

    # get the global y-limits for all axes
    if ylim is None:
        ylim = [np.floor(np.amin(colds) / 100) * 100, np.ceil(np.amax(colds) / 100) * 100]

    # water
    axs[0].set_ylabel('Bq/mL')
    axs[0].plot(Cntd['sinsrt'], r_h2o, line_style, color=colour)
    axs[0].set_ylim(ylim)
    axs[0].set_xlim([0, 25])
    axs[0].set_title('water')

    if draw_bckg:
        rect = patches.Rectangle((0, 0), 12.5, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor='b', alpha=0.1)
        # add the patch to the plot
        axs[0].add_patch(rect)
        rect = patches.Rectangle((12.5, 0), 1.5, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor=str(0.8))
        # add the patch to the plot
        axs[0].add_patch(rect)
        rect = patches.Rectangle((14, 0), 11, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor='red', alpha=0.1)
        # add the patch to the plot
        axs[0].add_patch(rect)

    # air
    axs[1].plot(Cntd['sinsrt'], r_air, line_style, color=colour)
    axs[1].set_ylim(ylim)
    axs[1].set_xlim([0, 25])
    axs[1].get_yaxis().set_ticks([])
    axs[1].set_title('air')
    axs[1].set_xlabel('ring radii [mm]')

    if draw_bckg:
        rect = patches.Rectangle((0, 0), 12.5, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor='None')
        # add the patch to the plot
        axs[1].add_patch(rect)
        rect = patches.Rectangle((12.5, 0), 1.5, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor=str(0.8))
        # add the patch to the plot
        axs[1].add_patch(rect)
        rect = patches.Rectangle((14, 0), 11, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor='red', alpha=0.1)
        # add the patch to the plot
        axs[1].add_patch(rect)

    # bone
    axs[2].plot(Cntd['sinsrt'], r_bone, line_style, color=colour)
    axs[2].set_ylim(ylim)
    axs[2].set_xlim([0, 25])
    axs[2].get_yaxis().set_ticks([])
    axs[2].set_title('bone')

    if draw_bckg:
        rect = patches.Rectangle((0, 0), 12.5, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor='0.85')
        axs[2].add_patch(rect)
        rect = patches.Rectangle((12.5, 0), 12.5, ylim[1], linewidth=0, edgecolor='None',
                                 facecolor='red', alpha=0.1)
        axs[2].add_patch(rect)

    return axs


def plot_uniformity(im, Cntd, masks, inserts_area=True, uniform_area=True, ylim=None,
                    ring_draw_ymin=None, ring_drw_ymax=None, grid=False, axis=None):
    """plot the uniformity regions using 18 sampling rings"""
    if axis is None:
        _, ax = plt.subplots(1)
    else:
        ax = axis

    # background around inserts
    if inserts_area:
        r_bi = extract_rings(im, masks['fst_ibckg'], l0=200, l1=218)
        plt.plot(Cntd['sbckgs'], r_bi, '.-', color='k', label='insert area')

    # background in the middle uniform part
    if uniform_area:
        r_b = extract_rings(im, masks['fst_insrt'], l0=300, l1=318)
        plt.plot(Cntd['sbckgs'], r_b, '.--', color='k', label='uniform area')

    plt.legend()

    if ylim is not None:
        plt.ylim(ylim)

    plt.ylabel('Bq/mL')
    plt.xlabel('sampling ring radii [mm]')
    plt.title('background rings')

    if ring_drw_ymax is not None and ring_draw_ymin is not None:
        for k in range(len(Cntd['rbckgs'])):
            rect = patches.Rectangle((Cntd['rbckgs'][k] - 6, ring_draw_ymin), 6, ring_drw_ymax,
                                     linewidth=0, edgecolor='None',
                                     facecolor=str(0.8 + (k%2) * 0.1))
            # add the patch to the plot
            ax.add_patch(rect)

    if grid: plt.grid('on')
    plt.show()


def plot_rods(Cntd, masks, im, color='k', ylim=None, pick_rods=None, draw_rods=True,
              contrast_ref=None, line_style='.-', out_raw=False):
    """
    Calculate the resolution curves for each set of rod diameters.
    Return the contrast for each rod set.
    """

    # pick the rods for which to plot the recovery
    if pick_rods is None:
        pick_rods = range(0, len(Cntd['rods_nrngs']))

    # contrast and ratio values for output
    cntrst = np.zeros(len(pick_rods))
    ratios = np.zeros(len(pick_rods))

    # raw values for each ring and selected rod
    rawval = np.zeros((len(pick_rods), np.max(Cntd['rods_nrngs'][pick_rods])), dtype=np.float32)

    # sampling
    for k, (nrng, off, rad) in enumerate(
            zip(Cntd['rods_nrngs'][pick_rods], Cntd['rods_off'][pick_rods],
                Cntd['rods_rad'][pick_rods])):

        print(f'# sampling rings {nrng} with offset {off} and radius {rad}')

        vrng = np.zeros(nrng, dtype=np.float32)
        for i in range(nrng):
            msk = masks['fst_res'] == i + off
            vrng[i] = np.sum(im[msk]) / np.sum(msk)

        rawval[k, :nrng] = vrng

        ratios[k] = vrng[-1] / vrng[0]
        if contrast_ref is None:
            cntrst[k] = (vrng[-1] - vrng[0]) / vrng[-1]
        else:
            cntrst[k] = (vrng[-1] - vrng[0]) / contrast_ref

        plt.plot(Cntd['rods_rngc'][:nrng], vrng, line_style, color=color)

    if draw_rods:
        for rx in Cntd['rods_rad'][pick_rods]:
            plt.axvline(x=rx, color='0.8', linestyle='-', lw=1)

    plt.ylabel('Bq/mL')
    plt.xlabel('sampling ring radii [mm]')
    plt.title('ACR resolution rods')
    plt.show()

    return rawval if out_raw else (cntrst, ratios)



import logging

import numpy as np
import scipy.ndimage as ndi

try:
    import cuvec as cu

    from . import improc
except ImportError: # GPU routines not compiled
    cu, improc = None, None

__all__ = ['conv_separable', 'isub', 'nlm']
log = logging.getLogger(__name__)
FLOAT_MAX = np.float32(np.inf)


def conv_separable(vol, knl, dev_id=0, output=None, sync=True):
    """
    Args:
      vol(ndarray): Can be any number of dimensions `ndim`
        (GPU requires `ndim <= 3`).
      knl(ndarray): `ndim` x `width` separable kernel
        (GPU requires `width <= 17`).
      dev_id(int or bool): GPU device ID to try [default: 0].
        Set to `False` to force CPU fallback.
    """
    assert vol.ndim == len(knl)
    assert knl.ndim == 2
    if len(knl) > 3 or knl.shape[1] > 17:
        log.warning("kernel larger than 3 x 17 not supported on GPU")
        dev_id = False
    if improc is not None and dev_id is not False:
        log.debug("GPU conv")

        pad = 3 - len(knl)        # <3 dims
        k_pad = 17 - knl.shape[1] # kernel width < 17

        if pad or k_pad:
            knl = np.pad(knl, [(0, pad), (k_pad // 2, k_pad//2 + (k_pad%2))])
            if pad:
                knl[-pad:, 17 // 2] = 1
                vol = vol.reshape(vol.shape + (1,) * pad)
        src = cu.asarray(vol, 'float32')
        knl = cu.asarray(knl, 'float32')
        if output is not None:
            output = cu.asarray(output, 'float32')
        dst = improc.convolve(src, knl, output=output, dev_id=dev_id, sync=sync,
                              log=log.getEffectiveLevel())
        res = cu.asarray(dst, vol.dtype)
        return res[(slice(0, None),) * (res.ndim - pad) + (-1,) * pad] if pad else res
    else:
        log.debug("CPU conv")
        for dim in range(len(knl)):
            h = knl[dim].reshape((1,) * dim + (-1,) + (1,) * (len(knl) - dim - 1))
            vol = ndi.convolve(vol, h, output=output, mode='constant', cval=0.)
        return vol


def check_cuvec(a, shape, dtype, allow_none=True):
    """Asserts that CuVec `a` is of `shape` and `dtype`"""
    if a is None:
        assert allow_none, "must not be None"
        return
    if not isinstance(a, cu.CuVec):
        raise TypeError("must be a CuVec")
    if np.dtype(a.dtype) != np.dtype(dtype):
        raise TypeError(f"dtype must be {dtype}: got {a.dtype}")
    if a.shape != shape:
        raise IndexError(f"shape must be {shape}: got {a.shape}")


def nlm(img, ref, sigma=1, half_width=4, output=None, dev_id=0, sync=True):
    """
    3D Non-local means (NLM) guided filter.
    Args:
      img(3darray): input image to be filtered.
      ref(3darray): reference (guidance) image.
      sigma(float): NLM parameter.
      half_width(int): neighbourhood half-width.
      output(CuVec): pre-existing output memory.
      sync(bool): whether to `cudaDeviceSynchronize()` after GPU operations.
    Reference: https://doi.org/10.1109/CVPR.2005.38
    """
    img = cu.asarray(img, 'float32')
    ref = cu.asarray(ref, 'float32')
    if img.shape != ref.shape:
        raise IndexError(f"{img.shape} and {ref.shape} don't match")
    if img.ndim != 3:
        raise IndexError(f"must be 3D: got {img.ndim}D")
    check_cuvec(output, img.shape, 'float32')
    return cu.asarray(
        improc.nlm(img, ref, sigma=sigma, half_width=half_width, output=output, dev_id=dev_id,
                   sync=sync, log=log.getEffectiveLevel()))


def isub(img, idxs, output=None, dev_id=0, sync=True):
    """
    output = img[idxs, ...]
    Args:
      img(2darray): input image.
      idxs(1darray['int32']): indicies into the first dimension of `img`.
      output(CuVec): pre-existing output memory.
      sync(bool): whether to `cudaDeviceSynchronize()` after GPU operations.
    """
    img = cu.asarray(img, 'float32')
    idxs = cu.asarray(idxs, 'int32')
    if img.ndim != 2:
        raise IndexError(f"must be 2D: got {img.ndim}D")
    check_cuvec(output, (idxs.shape[0], img.shape[1]), 'float32')
    return cu.asarray(
        improc.isub(img, idxs, output=output, dev_id=dev_id, sync=sync,
                    log=log.getEffectiveLevel()))



"""ACR/Jaszczak PET phantom design, image reconstruction and analysis"""
__author__ = "Pawel Markiewicz"
__copyright__ = "Copyright 2021-3"

from pathlib import Path

import numpy as np
import scipy
from scipy.optimize import curve_fit

from ..prc import imio, prc


def erf(x, a, u, k, b):
    """error function"""
    return a * scipy.special.erf(k * (x-u)) + b


def derf(x, a, u, k):
    """derivative of error function"""
    return a * k * 2 / (np.pi**.5) * np.exp(-(k * (x-u))**2)


def standard_analysis(
        fim,
        vois,
        fwhm=4.,
        patient_weight=70,  # kg
        simulated_dose=220, # MBq
        width_mm=10,
        zoffset=0):
    """
    Perform the standard ACR analysis.
    Arguments:
      fim:        the input NIfTI image in high resolution.
      vois:       the masks for VOIs to perform the analysis.
      fwhm:       the FWHM of the smoothing Gaussian kernel.
      zoffset:    offset from the middle of the axial extension
                  of the inserts.  By default the slice for
                  the analysis is in the middle of the insert's
                  axial extension.
      patient_does: patient simulated does in MBq
      patient_weight: patient weight in kg
    """

    fim = Path(fim)
    if not fim.is_file():
        raise IOError('unrecognised input NIfTI file')

    imd = imio.getnii(fim, output='all')

    # > for SUV calculations
    dose2wght = simulated_dose * 1e6 / (patient_weight*1e3)

    im_smo = prc.imsmooth(imd['im'], fwhm=fwhm, voxsize=imd['voxsize'])

    im_suv = im_smo / dose2wght

    # > axial width for standard analysis
    width_vox = int(np.round(width_mm / imd['voxsize'][0]))

    # > axial voxel range
    zrng = vois['r_insrt']

    z_start_idx = int(np.mean(zrng) - width_vox/2)

    msk = np.zeros(vois['s_i1'].shape, dtype=bool)

    zi = z_start_idx + zoffset
    msk[zi:zi + width_vox, ...] = True

    # > (A) Contrast
    h1 = np.float32(np.max(im_suv[vois['s_i1'] * msk]))
    h2 = np.max(im_suv[vois['s_i2'] * msk])
    h3 = np.max(im_suv[vois['s_i3'] * msk])
    h4 = np.max(im_suv[vois['s_i4'] * msk])

    # > (B) Scatter/Attenuation
    bckg_avg = np.float32(np.mean(im_suv[vois['s_bckg'] * msk]))
    bone_avg = np.mean(im_suv[vois['s_b'] * msk])
    h2o_avg = np.mean(im_suv[vois['s_w'] * msk])
    air_avg = np.mean(im_suv[vois['s_a'] * msk])

    bckg_min = np.min(im_suv[vois['s_bckg'] * msk])
    bone_min = np.min(im_suv[vois['s_b'] * msk])
    h2o_min = np.min(im_suv[vois['s_w'] * msk])
    air_min = np.min(im_suv[vois['s_a'] * msk])

    # > Ratio Calculations
    h1_bckg = h1 / bckg_avg
    h2_bckg = h2 / bckg_avg
    h3_bckg = h3 / bckg_avg
    h4_bckg = h4 / bckg_avg

    h2_h1 = np.float32(h2 / h1)
    h3_h1 = h3 / h1
    h4_h1 = h4 / h1

    air_bone = air_min / bone_min
    h2o_bone = h2o_min / bone_min

    out = {
        'h1max': h1, 'h2max': h2, 'h3max': h3, 'h4max': h4, 'bckg_avg': bckg_avg,
        'bone_avg': bone_avg, 'h2o_avg': h2o_avg, 'air_avg': air_avg, 'bckg_min': bckg_min,
        'bone_min': bone_min, 'h2o_min': h2o_min, 'air_min': air_min, 'h1_bckg': h1_bckg,
        'h2_bckg': h2_bckg, 'h3_bckg': h3_bckg, 'h4_bckg': h4_bckg, 'h2_h1': h2_h1, 'h3_h1': h3_h1,
        'h4_h1': h4_h1, 'air_bone': air_bone, 'h2o_bone': h2o_bone}

    for k in out:
        out[k] = np.float32(out[k])

    test1 = bckg_avg > 0.85 and bckg_avg < 1.15
    test2 = h1 > 1.8 and h1 < 2.8
    test3 = h2_h1 > 0.7

    out['test_bckg_avg'] = 'PASS'*test1 + 'FAIL' * ~test1 + ': ' + str(round(bckg_avg,
                                                                             2)) + ' (0.85-1.15)'
    out['test_25mm_insert'] = 'PASS'*test2 + 'FAIL' * ~test2 + ': ' + str(round(h1,
                                                                                2)) + ' (1.8-2.8)'
    out['test_16/25_insert'] = 'PASS'*test3 + 'FAIL' * ~test3 + ': ' + str(round(h2_h1,
                                                                                 2)) + ' (>0.7)'

    return out


def estimate_fwhm(fim, vois, Cntd, insert='water'):
    ''' Estimate the effective image resolution
        for any given ACR cylindrical insert
    '''

    from matplotlib import pyplot as plt

    if isinstance(fim, (str, Path)) and Path(fim).is_file():
        im = imio.getnii(fim)
    elif isinstance(fim, np.ndarray) and fim.ndim == 3:
        im = fim
    else:
        raise IOError('unrecognised input file or Numpy array')

    if im.shape != vois['fst_insrt'].shape:
        raise ValueError('the VOIs shape is incompatible with the image')

    ins = insert

    # > ring index ranges
    RIR = {
        'bone': [90, 102], # bone
        'air': [70, 82],   # air
        'water': [50, 62], # water
        'hot1': [10, 20],  # from biggest to smallest hot insert
        'hot2': [20, 30],
        'hot3': [30, 40],
        'hot4': [40, 50]}

    if ins not in RIR:
        raise ValueError('unrecognised insert name')

    # > pick what template to use (reduced axially or full)
    tmplt = vois['fst_insrt']
    tmplt3 = vois['fst_insrt3']

    # > sampling for analytical edge functions (transaxial)
    x = np.linspace(-1, np.max(Cntd['sinsrt']) + 1)

    # > extracted ring index values
    riv = np.zeros((RIR[ins][1] - RIR[ins][0]), dtype=np.float64)

    res = {
        'y': np.zeros(len(x), dtype=np.float64), 'dy': np.zeros(len(x), dtype=np.float64),
        'fwhm': 0, 'peak': 0, 'parg': 0, 'mnmx': 0}

    irngs = range(RIR[ins][0], RIR[ins][1])
    nrng = len(irngs)

    # > extract rings
    for i, l in enumerate(irngs):
        if ins == 'hot3':
            msk = tmplt3 == l
        else:
            msk = tmplt == l
        riv[i] = np.sum(im[msk]) / np.sum(msk)

    r = Cntd['sinsrt'][range(nrng)]
    re = x

    # ------- ERF fitting and derivative --------
    pe0 = [np.max(riv), np.median(r), .1, np.min(riv)]

    pe, pcov = curve_fit(erf, r, riv, pe0)

    ye = erf(re, *pe)
    res['y'] = ye

    dye = derf(re, *pe[:-1])
    res['dy'] = dye

    sgm = 1 / (np.sqrt(2) * pe[2])
    res['fwhm'] = 2 * np.sqrt(2 * np.log(2)) * sgm
    res['peak'] = 2 / (np.pi**.5) * pe[2] * pe[0]

    # > inflection point (should be around the insert border/wall)
    res['parg'] = pe[1]

    if ins[0] == 'h':
        res['mnmx'] = np.max(riv)
        res['dy'] *= -1
    else:
        res['mnmx'] = np.min(riv)
    res['r'] = r

    fig, ax = plt.subplots(2, 1)
    ax[0].plot(res['r'], riv, 'o')
    ax[0].plot(x, res['y'])
    ax[1].plot(x, res['dy'])
    ax[0].set_title('Insert {}: FWHM = {} mm'.format(ins, round(res['fwhm'], 2)))
    ax[0].set_ylabel('Bq/ML')
    ax[1].set_ylabel('Bq/ML')
    ax[1].set_xlabel('distance from insert centre [mm]')

    res['r_values'] = riv
    return res



"""ACR/Jaszczak PET phantom I/O and auxiliary functions"""
__author__ = "Pawel Markiewicz"
__copyright__ = "Copyright 2021-23"

import os
from itertools import chain
from pathlib import Path, PurePath
from subprocess import run

import dcm2niix
from miutil.fdio import hasext

from ..prc import imio, prc


def preproc(indat, Cntd, smooth=True, reftrim='', outpath=None, mode='nac'):
    """Convert to NIfTI (if DICOM), smooth using the Gaussian and trim/scale up."""
    opth = Path(indat).parent if outpath is None else Path(outpath)

    if mode == 'nac':
        outdir = opth / mode.upper()
    elif mode[:3] == 'qnt':
        outdir = opth / mode.upper()
    else:
        raise ValueError('unrecognised mode')

    imio.create_dir(outdir)

    if isinstance(indat, (str, PurePath)) and Path(indat).is_dir():
        # CONVERT TO NIfTI
        if not imio.dcmdir(indat):
            raise IOError('the provided folder does not contain DICOM files')
        for f in chain(outdir.glob('*.nii*'), outdir.glob('*.json')):
            # remove previous files
            os.remove(f)
        run([dcm2niix.bin, '-i', 'y', '-v', 'n', '-o', outdir, '-f', '%f_%s', str(indat)])
        fnii = list(outdir.glob('*offline3D*.nii*'))
        if len(fnii) == 1:
            fnii = fnii[0]
        else:
            raise ValueError('Confusing or missing NIfTI output')
    elif isinstance(indat, (str, PurePath)) and Path(indat).is_file() and hasext(
            indat, ('nii', 'nii.gz')):
        fnii = Path(indat)
    else:
        raise ValueError('the input NIfTI file or DICOM folder do not exist')

    # > Gaussian smooth image data if needed
    if smooth:
        if Cntd['fwhm_' + mode[:3]] > 0:
            smostr = '_smo-' + str(Cntd['fwhm_' + mode]).replace('.', '-') + 'mm'
            fnii = prc.imsmooth(
                fnii, fwhm=Cntd['fwhm_' + mode[:3]], fout=outdir /
                (mode.upper() + '_' + fnii.name.split('.nii')[0] + smostr + '.nii.gz'),
                output='file')

    Cntd['f' + mode] = fnii

    # > trim and upsample the PET
    imup = prc.imtrimup(
        fnii,
        refim=reftrim,
        scale=Cntd['sclt'],
        int_order=Cntd['interp'],
        fmax=0.1,                                       # controls how much trimming there is
        fcomment_pfx=fnii.name.split('.nii')[0] + '__',
        store_img=True)

    Cntd[f'f{mode}up'] = Path(imup['fim'])

    return Cntd



"""image input/output functionalities."""
import datetime
import logging
import numbers
import os
import pathlib
import re
import shutil
from pathlib import Path
from subprocess import run
from textwrap import dedent

import nibabel as nib
import numpy as np
import pydicom as dcm
from miutil.fdio import create_dir, hasext
from miutil.imio.nii import getnii  # NOQA: F401 # yapf: disable
from miutil.imio.nii import nii_gzip  # NOQA: F401 # yapf: disable
from miutil.imio.nii import nii_ugzip  # NOQA: F401 # yapf: disable
from miutil.imio.nii import niisort  # NOQA: F401 # yapf: disable
from miutil.imio.nii import array2nii

# > NiftyPET resources
from .. import resources as rs

log = logging.getLogger(__name__)

# > possible extensions for DICOM files
dcmext = 'dcm', 'ima', 'img'

# > remove characters unwanted in file/folder names
avoid_chars = '/{}[]!@#$%^&*.()+=:;~ '

# > DICOM coding of PET isotopes
istp_code = {
    'C-111A1': 'F18', 'C-105A1': 'C11', 'C-B1038': 'O15', 'C-128A2': 'Ge68', 'C-131A3': 'Ga68'}


def time_stamp(simple_ascii=False):
    now = datetime.datetime.now()
    if simple_ascii:
        nowstr = str(now.year) + '-' + str(now.month) + '-' + str(now.day) + '_' + str(
            now.hour) + 'h' + str(now.minute)
    else:
        nowstr = str(now.year) + '-' + str(now.month) + '-' + str(now.day) + ' ' + str(
            now.hour) + ':' + str(now.minute)
    return nowstr


def fwhm2sig(fwhm, voxsize=2.0):
    return (fwhm/voxsize) / (2 * (2 * np.log(2))**.5)


def rem_chars(txt, replacement_char='_'):
    ''' remove disallowed or inconvenient characters
        (as def. in `avoid_chars`) from file/folder names.
    '''
    for c in avoid_chars:
        txt = txt.replace(c, '_')
    txt = txt.lstrip(replacement_char)
    return txt or 'undefined'


def isdcm(f):
    try:
        dcm.dcmread(f)
    except Exception:
        return False
    else:
        return True


def dcmdir(inpth):
    '''
    Check if the folder has DICOM files and
    specify them
    '''
    if not inpth.is_dir():
        raise IOError('unrecognised folder')
    dcmlst = [f for f in inpth.iterdir() if isdcm(f)]
    Ndcm = len(dcmlst) # number of DICOM files
    return {'fdcm': dcmlst, 'N': Ndcm} if Ndcm > 0 else None


def mgh2nii(fim, fout=None, output=None):
    '''
    Convert `*.mgh` or `*.mgz` FreeSurfer image to NIfTI.
        Arguments:
            fim: path to the input MGH file
            fout: path to the output NIfTI file, if None then
                  creates based on `fim`
            output: if not None and an applicable string it will
            output a dictionary or an array (see below)
        Return:
            None: returns nothing
            'image' or 'im': outputs just the image
            'affine': outputs just the affine matrix
            'all': outputs all as a dictionary
    '''

    if not os.path.isfile(fim):
        raise ValueError('The input path is incorrect!')

    # > get the image dictionary
    mghd = getmgh(fim, output='all')
    im = mghd['im']

    # > sort out the output
    if fout is None:
        fout = fim.parent / (fim.name.split('.')[0] + '.nii.gz')

    out = fout
    if output == 'image' or output == 'im':
        out = fout, im
    elif output == 'affine':
        out = fout, mghd['affine']
    elif output == 'all':
        out = mghd
        out['fout'] = fout

    array2nii(
        mghd['im'], mghd['affine'], fout,
        trnsp=(mghd['transpose'].index(0), mghd['transpose'].index(1), mghd['transpose'].index(2)),
        flip=mghd['flip'])

    return out


def getmgh(fim, nan_replace=None, output='image'):
    '''
    Get image from `*.mgz` or `*.mgh` file (FreeSurfer).
    Arguments:
        fim: input file name for the MGH/Z image
        output: option for choosing output: 'image', 'affine' matrix or
                'all' for a dictionary with all the info.
    Return:
        'image': outputs just the image
        'affine': outputs just the affine matrix
        'all': outputs all as a dictionary
    '''

    if not os.path.isfile(fim):
        raise ValueError('The input path is incorrect!')

    mgh = nib.freesurfer.load(str(fim))

    if output == 'image' or output == 'all':

        imr = np.asanyarray(mgh.dataobj)
        # replace NaNs if requested
        if isinstance(nan_replace, numbers.Number):
            imr[np.isnan(imr)] = nan_replace

        imr = np.squeeze(imr)
        dimno = imr.ndim

        # > get orientations from the affine
        ornt = nib.io_orientation(mgh.affine)
        trnsp = tuple(np.flip(np.argsort(ornt[:, 0])))
        flip = tuple(np.int8(ornt[:, 1]))

        # > flip y-axis and z-axis and then transpose
        if dimno == 4:   # dynamic
            imr = np.transpose(imr[::-flip[0], ::-flip[1], ::-flip[2], :], (3,) + trnsp)
        elif dimno == 3: # static
            imr = np.transpose(imr[::-flip[0], ::-flip[1], ::-flip[2]], trnsp)

        # # > voxel size
        # voxsize = mgh.header.get('pixdim')[1:mgh.header.get('dim')[0] + 1]
        # # > rearrange voxel size according to the orientation
        # voxsize = voxsize[np.array(trnsp)]

    if output == 'all':
        out = {
            'im': imr, 'affine': mgh.affine, 'fim': fim, 'dtype': mgh.get_data_dtype(),
            'shape': imr.shape, 'hdr': mgh.header, 'transpose': trnsp, 'flip': flip}
    elif output == 'image':
        out = imr
    elif output == 'affine':
        out = mgh.affine
    else:
        raise NameError("Unrecognised output request!")

    return out


def getnii_descr(fim):
    '''Extracts the custom description header field to dictionary'''
    nim = nib.load(fim)
    hdr = nim.header
    rcnlst = hdr['descrip'].item().split(';')
    rcndic = {}

    if rcnlst[0] == '':
        return rcndic

    for ci in range(len(rcnlst)):
        tmp = rcnlst[ci].split('=')
        rcndic[tmp[0]] = tmp[1]
    return rcndic


def orientnii(imfile, Cnt=None):
    '''Get the orientation from NIfTI sform.  Not fully functional yet.'''

    # > check if the dictionary of constant is given
    if Cnt is None:
        Cnt = {}

    strorient = ['L-R', 'S-I', 'A-P']
    niiorient = []
    niixyz = np.zeros(3, dtype=np.int8)
    if os.path.isfile(imfile):
        nim = nib.load(imfile)
        A = nim.get_sform()
        for i in range(3):
            niixyz[i] = np.argmax(abs(A[i, :-1]))
            niiorient.append(strorient[niixyz[i]])
        log.info('NIfTI orientation:\n{}'.format(niiorient))

    return niiorient


def pick_t1w(mri):
    '''Pick the MR T1w from the dictionary for MR->PET registration.'''
    if not isinstance(mri, dict):
        raise IOError('incorrect input given for the T1w image')

    # check if NIfTI file is given
    if 'T1N4' in mri and os.path.isfile(mri['T1N4']):
        return mri['T1N4']
    # or another bias corrected
    elif 'T1bc' in mri and os.path.isfile(mri['T1bc']):
        return mri['T1bc']
    elif 'T1nii' in mri and os.path.isfile(mri['T1nii']):
        return mri['T1nii']
    elif 'T1DCM' in mri and os.path.exists(mri['MRT1W']):
        return dcm2nii(mri['T1nii'], 'converted')
    raise IOError('could not find a T1w image!')


# ======================================================================
def dcminfo(dcmvar, Cnt=None, output='class', t1_name='mprage'):
    """
    Get DICOM info from file/header.
    Args:
      dcmvar: DICOM header as a file/string/dictionary
      Cnt(dict): constants used in advanced reconstruction or analysis
      output(str): 'class' outputs all with classification;
        'basic' outputs scanner ID and series/protocol string;
        'detail' gives most relevant tags in dictionary
      t1_name(str): helps identify T1w MR image present in series or file names
    """
    if Cnt is None:
        Cnt = {}

    if isinstance(dcmvar, str):
        log.debug('provided DICOM file: {}'.format(dcmvar))
        dhdr = dcm.dcmread(dcmvar)
    elif isinstance(dcmvar, dict):
        dhdr = dcmvar
    elif isinstance(dcmvar, dcm.dataset.FileDataset):
        dhdr = dcmvar
    elif isinstance(dcmvar, (pathlib.Path, pathlib.PurePath)):
        dhdr = dcm.dcmread(dcmvar)

    dtype = dhdr[0x08, 0x08].value
    log.debug('   Image Type: {}'.format(dtype))

    # > output dictionary
    outdct = {}

    # ------------------------------------------
    # > scanner ID
    scanner_vendor = 'unknown'
    if [0x008, 0x070] in dhdr:
        scanner_vendor = dhdr[0x008, 0x070].value
        outdct['scnr_vndr'] = scanner_vendor

    scanner_model = 'unknown'
    if [0x008, 0x1090] in dhdr:
        scanner_model = dhdr[0x008, 0x1090].value
        outdct['scnr_mdl'] = scanner_model

    scanner_id = 'other'
    if any(s in scanner_model
           for s in ['mMR', 'Biograph']) and 'siemens' in scanner_vendor.lower():
        scanner_id = 'mmr'
    elif 'signa' in scanner_model.lower() and 'ge' in scanner_vendor.lower():
        scanner_id = 'signa'

    outdct['scnr_id'] = scanner_id
    # ------------------------------------------

    # ------------------------------------------
    # > date/time
    study_time = None
    if [0x008, 0x030] in dhdr and [0x008, 0x020] in dhdr:
        val = dhdr[0x008, 0x020].value + dhdr[0x008, 0x030].value
        if '.' in val:
            study_time = datetime.datetime.strptime(val, '%Y%m%d%H%M%S.%f')
        else:
            study_time = datetime.datetime.strptime(val, '%Y%m%d%H%M%S')

        outdct['st_time'] = study_time

    series_time = None
    if [0x008, 0x031] in dhdr and [0x008, 0x021] in dhdr:
        val = dhdr[0x008, 0x021].value + dhdr[0x008, 0x031].value
        if '.' in val:
            series_time = datetime.datetime.strptime(val, '%Y%m%d%H%M%S.%f')
        else:
            series_time = datetime.datetime.strptime(val, '%Y%m%d%H%M%S')

        outdct['sr_time'] = series_time

    acq_time = None
    if [0x008, 0x032] in dhdr and [0x008, 0x022] in dhdr:
        val = dhdr[0x008, 0x022].value + dhdr[0x008, 0x032].value
        if '.' in val:
            acq_time = datetime.datetime.strptime(val, '%Y%m%d%H%M%S.%f')
        else:
            acq_time = datetime.datetime.strptime(val, '%Y%m%d%H%M%S')

        outdct['aq_time'] = acq_time

    # > frame duration time
    if [0x018, 0x1242] in dhdr:
        val = int(dhdr[0x018, 0x1242].value)
        outdct['frm_dur'] = datetime.timedelta(milliseconds=val)

    # ------------------------------------------

    # > institution
    inst = ''
    if [0x008, 0x080] in dhdr:
        inst = dhdr[0x008, 0x080].value
        outdct['inst_name'] = inst

    prtcl = ''
    if [0x18, 0x1030] in dhdr:
        prtcl = dhdr[0x18, 0x1030].value
        outdct['pr_name'] = prtcl

    srs = ''
    if [0x08, 0x103e] in dhdr:
        srs = dhdr[0x08, 0x103e].value
        outdct['sr_name'] = srs

    # > units
    unt = None
    if [0x054, 0x1001] in dhdr:
        unt = dhdr[0x054, 0x1001].value
        outdct['units'] = unt

    # > sequence name
    if [0x018, 0x024] in dhdr:
        outdct['sq_name'] = dhdr[0x018, 0x024].value

    # ------- PATIENT -------
    # > patient size
    if [0x010, 0x1020] in dhdr:
        outdct['pat_size'] = dhdr[0x010, 0x1020].value

    # > patient sex
    if [0x010, 0x040] in dhdr:
        outdct['pat_sex'] = dhdr[0x010, 0x040].value

    # > patient DoB
    if [0x010, 0x030] in dhdr:
        outdct['pat_dob'] = dhdr[0x010, 0x030].value

    # > patient weight
    if [0x010, 0x1030] in dhdr:
        outdct['pat_weight'] = dhdr[0x010, 0x1030].value

    # > patient age
    if [0x010, 0x1010] in dhdr:
        outdct['pat_age'] = dhdr[0x010, 0x1010].value
    # -----------------------

    # +++++++++++++++++++++++++++++++++++++++++++++
    if output == 'basic':
        out = [prtcl, srs, scanner_id]
        return out
    # +++++++++++++++++++++++++++++++++++++++++++++

    # =================================================================
    # PET
    # -----------------------------------------------------------------
    if [0x054, 0x1000] in dhdr or [0x054, 0x000] in dhdr:
        outdct['PET'] = {}

    # > PET parameters
    srs_type = None
    if [0x054, 0x1000] in dhdr:
        srs_type = dhdr[0x054, 0x1000].value[0]
        outdct['PET']['sr_type'] = srs_type

    recon = None
    if [0x054, 0x1103] in dhdr:
        recon = dhdr[0x054, 0x1103].value
        outdct['PET']['recon'] = recon

    decay_corr = None
    if [0x054, 0x1102] in dhdr:
        decay_corr = dhdr[0x054, 0x1102].value
        outdct['PET']['decay_crr'] = decay_corr

    # > decay factor
    dcf = None
    if [0x054, 0x1321] in dhdr:
        dcf = float(dhdr[0x054, 0x1321].value)
        outdct['PET']['decay_fcr'] = dcf

    atten = None
    if [0x054, 0x1101] in dhdr:
        atten = dhdr[0x054, 0x1101].value
        outdct['PET']['ac_mth'] = atten

    scat = None
    if [0x054, 0x1105] in dhdr:
        scat = dhdr[0x054, 0x1105].value
        outdct['PET']['sc_mth'] = scat

    # > scatter factor
    scf = None
    if [0x054, 0x1323] in dhdr:
        scf = float(dhdr[0x054, 0x1323].value)
        outdct['PET']['sc_fcr'] = scf

    # > randoms correction method
    rand = None
    if [0x054, 0x1100] in dhdr:
        rand = dhdr[0x054, 0x1100].value
        outdct['PET']['rc_mth'] = rand

    # > dose calibration factor
    dscf = None
    if [0x054, 0x1322] in dhdr:
        dscf = float(dhdr[0x054, 0x1322].value)
        outdct['PET']['dose_calib_fcr'] = dscf

    # > dead time factor
    dt = None
    if [0x054, 0x1324] in dhdr:
        dt = float(dhdr[0x054, 0x1324].value)
        outdct['PET']['dt_fcr'] = dt

    # -----------------------------------------------------------------
    # RADIO TRACER
    tracer = None
    tdose = None
    hlife = None
    pfract = None
    ttime0 = None
    ttime1 = None

    if [0x054, 0x016] in dhdr:
        # > all tracer info
        tinf = dhdr[0x054, 0x016][0]

        if [0x018, 0x031] in tinf:
            tracer = tinf[0x018, 0x031].value
            outdct['PET']['tracer'] = tracer

        if [0x018, 0x1074] in tinf:
            tdose = float(tinf[0x018, 0x1074].value)
            outdct['PET']['dose_tot'] = tdose

        if [0x018, 0x1075] in tinf:
            hlife = float(tinf[0x018, 0x1075].value)
            outdct['PET']['hlife'] = hlife

        if [0x018, 0x1076] in tinf:
            pfract = float(tinf[0x018, 0x1076].value)
            outdct['PET']['positron_frc'] = pfract

        if [0x018, 0x1078] in tinf:
            val = tinf[0x018, 0x1078].value
            if '.' in val:
                ttime0 = datetime.datetime.strptime(val, '%Y%m%d%H%M%S.%f')
            else:
                ttime0 = datetime.datetime.strptime(val, '%Y%m%d%H%M%S')

            outdct['PET']['radio_start_time'] = ttime0

        if [0x018, 0x1079] in tinf:
            val = tinf[0x018, 0x1079].value
            if '.' in val:
                ttime1 = datetime.datetime.strptime(val, '%Y%m%d%H%M%S.%f')
            else:
                ttime1 = datetime.datetime.strptime(val, '%Y%m%d%H%M%S')

            outdct['PET']['radio_stop_time'] = ttime1
    # -----------------------------------------------------------------

    # ---------------------------------------------
    # > MR parameters (echo time, etc)
    TR = None
    TE = None
    if [0x18, 0x80] in dhdr:
        TR = float(dhdr[0x18, 0x80].value)
        log.debug('   TR: {}'.format(TR))
    if [0x18, 0x81] in dhdr:
        TE = float(dhdr[0x18, 0x81].value)
        log.debug('   TE: {}'.format(TE))

    validTs = TR is not None and TE is not None

    if validTs:
        outdct['MR'] = {'TR': TR, 'TE': TE}
        """
        mrdct = {
            'series': srs, 'protocol': prtcl, 'units': unt, 'study_time': study_time, 'inst': inst,
            'series_time': series_time, 'acq_time': acq_time, 'scanner_id': scanner_id, 'TR': TR,
            'TE': TE}
        """
        mrdct = outdct
    # ---------------------------------------------

    # +++++++++++++++++++++++++++++++++++++++++++++
    if output == 'detail':
        return outdct
    # +++++++++++++++++++++++++++++++++++++++++++++

    isPET = (tracer is not None) and (srs_type in ['STATIC', 'DYNAMIC', 'WHOLE BODY', 'GATED'])

    # > CSA type (mMR)
    csatype = ''
    if [0x29, 0x1108] in dhdr:
        csatype = dhdr[0x29, 0x1108].value
        log.debug('   CSA Data Type: {}'.format(csatype))

    # > DICOM comment or on MR parameters
    cmmnt = ''
    if [0x20, 0x4000] in dhdr:
        cmmnt = dhdr[0x0020, 0x4000].value
        log.debug('   Comments: {}'.format(cmmnt))

    # > check for RAW data
    if any('PET_NORM' in s
           for s in dtype) or cmmnt == 'PET Normalization data' or csatype == 'MRPETNORM':
        out = ['raw', 'norm', scanner_id]

    elif any('PET_LISTMODE' in s
             for s in dtype) or cmmnt == 'Listmode' or csatype == 'MRPETLM_LARGE':
        out = ['raw', 'list', scanner_id]

    elif any('PET_EM_SINO' in s for s in dtype) or cmmnt == 'Sinogram' or csatype == 'MRPETSINO':
        out = ['raw', 'sinogram', scanner_id]

    # > physio data
    elif 'PET_PHYSIO' in dtype or 'physio' in cmmnt.lower():
        out = ['raw', 'physio', scanner_id]

    elif any('MRPET_UMAP3D' in s for s in dtype) or cmmnt == 'MR based umap':
        out = ['mr', 'mumap', 'ute', 'mr', scanner_id]

    elif isPET:
        """
        petdct = {
            'series': srs, 'protocol': prtcl, 'study_time': study_time, 'series_time': series_time,
            'inst': inst, 'acq_time': acq_time, 'scanner_id': scanner_id, 'type': srs_type,
            'units': unt, 'recon': recon, 'decay_corr': decay_corr, 'dcf': dcf,
            'attenuation': atten, 'scatter': scat, 'scf': scf, 'randoms': rand, 'dose_calib': dscf,
            'dead_time': dt, 'tracer': tracer, 'total_dose': tdose, 'half_life': hlife,
            'positron_fract': pfract, 'radio_start_time': ttime0, 'radio_stop_time': ttime1}
        """
        petdct = outdct

        out = ['pet', tracer.lower(), srs_type.lower(), scanner_id, petdct]

    # > a less stringent way of looking for the T1w sequence
    # > than the one below
    elif validTs and (t1_name in prtcl.lower() or t1_name in srs.lower()):
        out = ['mr', 't1', 'mprage', scanner_id, mrdct]

    elif validTs and TR > 400 and TR < 2500 and TE < 20:
        if t1_name in prtcl.lower() or t1_name in srs.lower():
            out = ['mr', 't1', 'mprage', scanner_id, mrdct]
        else:
            out = ['mr', 't1', scanner_id]

    elif validTs and TR > 2500 and TE > 50:
        out = ['mr', 't2', scanner_id, mrdct]

    # > UTE's two sequences:
    # > UTE2
    elif validTs and TR < 50 and TE < 20 and TE > 1:
        out = ['mr', 'ute', 'ute2', scanner_id, mrdct]

    # > UTE1
    elif validTs and TR < 50 and TE < 0.1 and TR > 0 and TE > 0:
        out = ['mr', 'ute', 'ute1', scanner_id, mrdct]

    else:
        out = ['unknown', str(cmmnt.lower())]

    return out


# ======================================================================


def dcmsort(folder, copy_series=False, Cnt=None, outpath=None, grouping='t+d'):
    '''
        sort out the DICOM files in the folder according to the recorded series.
        options:
        - copy_series: if True it makes folders for each series and copies DICOM
                       files into each corresponding folder.
        - Cnt: dictionary of constants (currently not used)
        - outpath: the output path where the series folders are created
        - grouping: defines how series are recognised, i.e., either by time plus
                    series description ('t+d') or by the description only ('d'),
                    or by acquisition and series times plus series description
                    ('a+t+d'), or using series instance unique id ('d+suid').
    '''

    # > insure that `folder` is Path object
    folder = Path(folder)
    if not folder.is_dir():
        raise ValueError('Incorrect input folder!')

    # > check if the dictionary of constant is given
    if Cnt is None:
        Cnt = {}

    srs = {}

    for f in folder.iterdir():

        # ---------------------------------
        if not f.is_file(): continue
        try:
            dhdr = dcm.dcmread(f)
        except Exception:
            continue
        # ---------------------------------

        # --------------------------------
        # > image size
        imsz = np.zeros(2, dtype=np.int64)
        if [0x028, 0x010] in dhdr:
            imsz[0] = dhdr[0x028, 0x010].value
        if [0x028, 0x011] in dhdr:
            imsz[1] = dhdr[0x028, 0x011].value

        # > voxel size
        vxsz = np.zeros(3, dtype=np.float64)
        if [0x028, 0x030] in dhdr and [0x018, 0x050] in dhdr:
            pxsz = np.array([float(e) for e in dhdr[0x028, 0x030].value])
            vxsz[:2] = pxsz
            vxsz[2] = float(dhdr[0x018, 0x050].value)

        # > orientation
        ornt = np.zeros(6, dtype=np.float64)
        if [0x020, 0x037] in dhdr:
            ornt = np.array([float(e) for e in dhdr[0x20, 0x37].value])

        # > series description, time and study time
        srs_dcrp = ''
        if [0x0008, 0x103e] in dhdr:
            srs_dcrp = dhdr[0x0008, 0x103e].value

        srs_uid = ''
        if [0x020, 0x00e] in dhdr:
            srs_uid = dhdr[0x020, 0x00e].value

        prtcl = ''
        # > protocol
        if [0x018, 0x1030] in dhdr:
            prtcl = dhdr[0x018, 0x1030].value

        # > series, study and acquisition times
        srs_time = dhdr[0x0008, 0x0031].value[:6]
        std_time = dhdr[0x0008, 0x0030].value[:6]
        acq_time = ''
        if [0x0008, 0x0032] in dhdr:
            acq_time = dhdr[0x0008, 0x0032].value[:6]

        # > study date
        std_date = dhdr[0x008, 0x020].value

        frm_dur = None
        # > frame duration time (for PET)
        if [0x018, 0x1242] in dhdr:
            val = np.round(int(dhdr[0x018, 0x1242].value) / 1e3, decimals=0)
            frm_dur = datetime.timedelta(seconds=val)

        # > time of tracer administration (start)
        tinjct = None
        # > PET tracer if present
        trcr = None
        if [0x054, 0x016] in dhdr:
            trinf = dhdr[0x054, 0x016][0]

            if [0x018, 0x1078] in trinf:
                val = trinf[0x018, 0x1078].value
                if '.' in val:
                    tinjct = datetime.datetime.strptime(val, '%Y%m%d%H%M%S.%f')
                else:
                    tinjct = datetime.datetime.strptime(val, '%Y%m%d%H%M%S')
            elif [0x018, 0x1072] in trinf:
                val = trinf[0x018, 0x1072].value
                if '.' in val:
                    tinjct = datetime.datetime.strptime(std_date + val, '%Y%m%d%H%M%S.%f')
                else:
                    tinjct = datetime.datetime.strptime(std_date + val, '%Y%m%d%H%M%S')

            if [0x018, 0x031] in trinf:
                trcr = trinf[0x018, 0x031].value
        # --------------------------------

        log.debug(
            dedent('''\
            --------------------------------------
            DICOM series desciption: {}
            DICOM series time: {}
            DICOM study  time: {}
            --------------------------------------''').format(srs_dcrp, srs_time, std_time))

        # ---------
        # series for any category (can be multiple scans within the same category)
        recognised_series = False
        srs_k = list(srs.keys())
        for s in srs_k:

            if grouping == 't+d':
                if (np.array_equal(srs[s]['imorient'], ornt)
                        and np.array_equal(srs[s]['imsize'], imsz)
                        and np.array_equal(srs[s]['voxsize'], vxsz)
                        and srs[s]['tseries'] == srs_time and srs[s]['series'] == srs_dcrp):
                    recognised_series = True
                    break

            elif grouping == 'a+t+d':
                if (np.array_equal(srs[s]['imorient'], ornt)
                        and np.array_equal(srs[s]['imsize'], imsz)
                        and np.array_equal(srs[s]['voxsize'], vxsz)
                        and srs[s]['tseries'] == srs_time and srs[s]['tacq'] == acq_time
                        and srs[s]['series'] == srs_dcrp):
                    recognised_series = True
                    break

            elif grouping == 'd':
                if (np.array_equal(srs[s]['imorient'], ornt)
                        and np.array_equal(srs[s]['imsize'], imsz)
                        and np.array_equal(srs[s]['voxsize'], vxsz)
                        and srs[s]['series'] == srs_dcrp):
                    recognised_series = True
                    break

            elif grouping == 'd+suid':
                if (np.array_equal(srs[s]['imorient'], ornt)
                        and np.array_equal(srs[s]['imsize'], imsz)
                        and np.array_equal(srs[s]['voxsize'], vxsz)
                        and srs[s]['series'] == srs_dcrp and srs[s]['series_uid'] == srs_uid):
                    recognised_series = True
                    break

            else:
                raise ValueError('Unrecognised grouping option')

        # >  if series was not found, create one
        if not recognised_series:
            if grouping == 't+d':
                s = srs_time + '_' + srs_dcrp
            elif grouping == 'a+t+d':
                s = acq_time + '_' + srs_time + '_' + srs_dcrp
            elif grouping == 'd':
                s = srs_dcrp
            elif grouping == 'd+suid':
                s = srs_dcrp + '_' + srs_uid[-10:]
            else:
                raise ValueError('Unrecognised grouping option')

            srs[s] = {}
            srs[s]['imorient'] = ornt
            srs[s]['imsize'] = imsz
            srs[s]['voxsize'] = vxsz
            srs[s]['tacq'] = acq_time
            srs[s]['tseries'] = srs_time
            srs[s]['tstudy'] = std_time
            srs[s]['dstudy'] = std_date
            srs[s]['series'] = srs_dcrp
            srs[s]['series_uid'] = srs_uid
            srs[s]['protocol'] = prtcl

            if tinjct is not None:
                srs[s]['radio_start_time'] = tinjct
            if frm_dur is not None:
                srs[s]['frm_dur'] = frm_dur
            if trcr is not None:
                srs[s]['tracer'] = trcr

        # append the file name
        srs[s].setdefault('files', [])

        if copy_series:
            out = folder
            if outpath is not None:
                try:
                    create_dir(outpath)
                except Exception as e:
                    log.warning(
                        f"could not create specified output folder, using input folder.\n\n{e}")
                else:
                    out = Path(outpath)

            srsdir = out / rem_chars(s)
            create_dir(srsdir)
            shutil.copy(f, srsdir)
            srs[s]['files'].append(srsdir / f.name)
        else:
            srs[s]['files'].append(f)

    return srs


# ======================================================================


def list_dcm_datain(datain):
    '''List all DICOM file paths in the datain dictionary of input data.'''

    if not isinstance(datain, dict):
        raise ValueError('The input is not a dictionary!')

    dcmlst = []
    # list of mu-map DICOM files
    if 'mumapDCM' in datain:
        dcmump = os.listdir(datain['mumapDCM'])
        # accept only *.dcm extensions
        dcmump = [os.path.join(datain['mumapDCM'], d) for d in dcmump if hasext(d, dcmext)]
        dcmlst += dcmump

    if 'T1DCM' in datain:
        dcmt1 = os.listdir(datain['T1DCM'])
        # accept only *.dcm extensions
        dcmt1 = [os.path.join(datain['T1DCM'], d) for d in dcmt1 if hasext(d, dcmext)]
        dcmlst += dcmt1

    if 'T2DCM' in datain:
        dcmt2 = os.listdir(datain['T2DCM'])
        # accept only *.dcm extensions
        dcmt2 = [os.path.join(datain['T2DCM'], d) for d in dcmt2 if hasext(d, dcmext)]
        dcmlst += dcmt2

    if 'UTE1' in datain:
        dcmute1 = os.listdir(datain['UTE1'])
        # accept only *.dcm extensions
        dcmute1 = [os.path.join(datain['UTE1'], d) for d in dcmute1 if hasext(d, dcmext)]
        dcmlst += dcmute1

    if 'UTE2' in datain:
        dcmute2 = os.listdir(datain['UTE2'])
        # accept only *.dcm extensions
        dcmute2 = [os.path.join(datain['UTE2'], d) for d in dcmute2 if hasext(d, dcmext)]
        dcmlst += dcmute2

    # list-mode data dcm
    if 'lm_dcm' in datain:
        dcmlst += [datain['lm_dcm']]

    if 'lm_ima' in datain:
        dcmlst += [datain['lm_ima']]

    # norm
    if 'nrm_dcm' in datain:
        dcmlst += [datain['nrm_dcm']]

    if 'nrm_ima' in datain:
        dcmlst += [datain['nrm_ima']]

    return dcmlst


def dcmanonym(dcmpth, displayonly=False, patient='anonymised', physician='anonymised',
              dob='19800101', Cnt=None):
    '''
    Anonymise DICOM file(s)
    Arguments:
        > dcmpth:   it can be passed as a single DICOM file, or
                    a folder containing DICOM files, or a list of DICOM file paths.
        > patient:  the name of the patient.
        > physician:the name of the referring physician.
        > dob:      patient's date of birth.
        > Cnt:      dictionary of constants (containing logging variable)
    '''
    # > check if the dictionary of constant is given
    if Cnt is None:
        Cnt = {}

    # > check if a single DICOM file
    if isinstance(dcmpth, str) and os.path.isfile(dcmpth):
        dcmlst = [dcmpth]
        log.debug('recognised the input argument as a single DICOM file.')

    # > check if a folder containing DICOM files
    elif isinstance(dcmpth, str) and os.path.isdir(dcmpth):
        dircontent = os.listdir(dcmpth)
        # > create a list of DICOM files inside the folder
        dcmlst = [
            os.path.join(dcmpth, d) for d in dircontent
            if os.path.isfile(os.path.join(dcmpth, d)) and hasext(d, dcmext)]
        log.debug('recognised the input argument as the folder containing DICOM files.')

    # > check if a folder containing DICOM files
    elif isinstance(dcmpth, list):
        if not all(os.path.isfile(d) and hasext(d, dcmext) for d in dcmpth):
            raise IOError('Not all files in the list are DICOM files.')
        dcmlst = dcmpth
        log.debug('recognised the input argument as the list of DICOM file paths.')

    # > check if dictionary of data input <datain>
    elif isinstance(dcmpth, dict) and 'corepath' in dcmpth:
        dcmlst = list_dcm_datain(dcmpth)
        log.debug('recognised the input argument as the dictionary of scanner data.')

    else:
        raise IOError('Unrecognised input!')

    for dcmf in dcmlst:
        # > read the file
        dhdr = dcm.dcmread(dcmf)

        # > get the basic info about the DICOM file
        dcmtype = dcminfo(dhdr)
        log.debug(
            dedent('''\
            --------------------------------------------------
            DICOM file is for: {}
            --------------------------------------------------''').format(dcmtype))

        # > anonymise mMR data.
        if 'mmr' in dcmtype:

            if [0x029, 0x1120] in dhdr and dhdr[0x029, 0x1120].name == '[CSA Series Header Info]':
                csafield = dhdr[0x029, 0x1120]
                csa = csafield.value
            elif [0x029, 0x1020] in dhdr and dhdr[0x029,
                                                  0x1020].name == '[CSA Series Header Info]':
                csafield = dhdr[0x029, 0x1020]
                csa = csafield.value
            else:
                csa = ''

            # string length considered for replacement
            strlen = 200

            idx = [m.start() for m in re.finditer(r'([Pp]atients{0,1}[Nn]ame)', csa)]
            if idx:
                log.debug(
                    'DICOM> found sensitive information deep in the headers: {}'.format(dcmtype))

            # > run the anonymisation
            iupdate = 0

            for i in idx:
                ci = i - iupdate

                if displayonly:
                    log.debug(
                        dedent('''\
                        DICOM> sensitive info:
                             {}''').format(csa[ci:ci + strlen]))
                    continue

                rplcmnt = re.sub(r'(\{\s*\"{1,2}\W*\w+\W*\w+\W*\"{1,2}\s*\})',
                                 '{ ""' + patient + '"" }', csa[ci:ci + strlen])
                # > update string
                csa = csa[:ci] + rplcmnt + csa[ci + strlen:]
                log.debug('DICOM> removed sensitive information.')
                # > correct for the number of removed letters
                iupdate = strlen - len(rplcmnt)

            # > update DICOM
            if not displayonly and csa != '':
                csafield.value = csa

        # > Patient's name
        if [0x010, 0x010] in dhdr:
            if displayonly:
                log.debug(
                    dedent('''\
                    DICOM> sensitive info: {}
                         > {}''').format(dhdr[0x010, 0x010].name, dhdr[0x010, 0x010].value))
            else:
                dhdr[0x010, 0x010].value = patient
                log.debug('DICOM> anonymised patients name')

        # > date of birth
        if [0x010, 0x030] in dhdr:
            if displayonly:
                log.debug(
                    dedent('''\
                DICOM> sensitive info: {}
                     > {}''').format(dhdr[0x010, 0x030].name, dhdr[0x010, 0x030].value))
            else:
                dhdr[0x010, 0x030].value = dob
                log.debug('   > anonymised date of birth')

        # > physician's name
        if [0x008, 0x090] in dhdr:
            if displayonly:
                log.debug(
                    dedent('''\
                DICOM> sensitive info: {}
                     > {}''').format(dhdr[0x008, 0x090].name, dhdr[0x008, 0x090].value))
            else:
                dhdr[0x008, 0x090].value = physician
                log.debug('   > anonymised physician name')

        dhdr.save_as(dcmf)


def dcm2nii(
    dcmpth,
    fimout='',
    fprefix='converted-from-DICOM_',
    fcomment='',
    outpath='',
    timestamp=True,
    executable=None,
    force=False,
):
    """
    Convert DICOM folder `dcmpth` to NIfTI using `dcm2niix` third-party software.
    Args:
      dcmpth: directory containing DICOM files
    """
    # skip conversion if the output already exists and not force is selected
    if os.path.isfile(fimout) and not force:
        return fimout

    if not executable:
        executable = getattr(rs, 'DCM2NIIX', None)
        if not executable:
            import dcm2niix
            executable = dcm2niix.bin
    if not os.path.isfile(executable):
        raise IOError(f"executable not found:{executable}")

    if not os.path.isdir(dcmpth):
        raise IOError("the provided `dcmpth` is not a folder")

    # output path
    if not outpath and fimout and '/' in fimout:
        opth = os.path.dirname(fimout)
        opth = opth or dcmpth
        fimout = os.path.basename(fimout)
    else:
        opth = outpath or dcmpth

    create_dir(opth)

    if not fimout:
        fimout = fprefix
        if timestamp:
            fimout += time_stamp(simple_ascii=True)
    fimout = fimout.split('.nii')[0]

    run([executable, '-f', fimout, '-o', opth, dcmpth])
    fniiout = list(pathlib.Path(opth).glob(f"*{fimout}*.nii*"))

    if fniiout:
        return str(fniiout[0])
    raise ValueError("could not find output nii file")


def dcm2im(fpth):
    '''
    Get the DICOM files from 'fpth' into an image with the affine transformation.
    fpth can be a list of DICOM files or a path (string) to the folder with DICOM files.
    '''
    # possible DICOM file extensions
    ext = 'dcm', 'ima'

    # case when given a folder path
    if isinstance(fpth, str) and os.path.isdir(fpth):
        SZ0 = len([d for d in os.listdir(fpth) if hasext(d, ext)])
        # list of DICOM files
        fdcms = os.listdir(fpth)
        fdcms = [os.path.join(fpth, f) for f in fdcms if hasext(f, ext)]

    # case when list of DICOM files is given
    elif isinstance(fpth, list) and os.path.isfile(os.path.join(fpth[0])):
        SZ0 = len(fpth)
        # list of DICOM files
        fdcms = fpth
        fdcms = [f for f in fdcms if hasext(f, ext)]
    else:
        raise NameError('Unrecognised input for DICOM files.')

    if SZ0 < 1:
        log.error('No DICOM images in the specified path.')
        raise IOError('Input DICOM images not recognised')

    # pick single DICOM header
    dhdr = dcm.read_file(fdcms[0])

    # -----------------------------------
    # some info, e.g.: patient position and series UID
    if [0x018, 0x5100] in dhdr:
        ornt = dhdr[0x18, 0x5100].value
    else:
        ornt = 'unkonwn'
    # Series UID
    sruid = dhdr[0x0020, 0x000e].value
    # -----------------------------------

    # -----------------------------------
    # INIT
    # image position
    P = np.zeros((SZ0, 3), dtype=np.float64)
    # image orientation
    Orn = np.zeros((SZ0, 6), dtype=np.float64)
    # xy resolution
    R = np.zeros((SZ0, 2), dtype=np.float64)
    # slice thickness
    S = np.zeros((SZ0, 1), dtype=np.float64)
    # slope and intercept
    SI = np.ones((SZ0, 2), dtype=np.float64)
    SI[:, 1] = 0

    # image data as an list of array for now
    IM = []
    # -----------------------------------

    c = 0
    for d in fdcms:
        dhdr = dcm.read_file(d)
        if [0x20, 0x32] in dhdr and [0x20, 0x37] in dhdr and [0x28, 0x30] in dhdr:
            P[c, :] = np.array([float(f) for f in dhdr[0x20, 0x32].value])
            Orn[c, :] = np.array([float(f) for f in dhdr[0x20, 0x37].value])
            R[c, :] = np.array([float(f) for f in dhdr[0x28, 0x30].value])
            S[c, :] = float(dhdr[0x18, 0x50].value)
        else:
            log.error('could not read all the DICOM tags.')
            return {'im': [], 'affine': [], 'shape': [], 'orient': ornt, 'sruid': sruid}

        if [0x28, 0x1053] in dhdr and [0x28, 0x1052] in dhdr:
            SI[c, 0] = float(dhdr[0x28, 0x1053].value)
            SI[c, 1] = float(dhdr[0x28, 0x1052].value)
        IM.append(dhdr.pixel_array)
        c += 1

    # check if orientation/resolution is the same for all slices
    if np.sum(Orn - Orn[0, :]) > 1e-6:
        log.error('varying orientation for slices')
    else:
        Orn = Orn[0, :]
    if np.sum(R - R[0, :]) > 1e-6:
        log.error('varying resolution for slices')
    else:
        R = R[0, :]

    # Patient Position
    # Rows and Columns
    if [0x28, 0x10] in dhdr and [0x28, 0x11] in dhdr:
        SZ2 = dhdr[0x28, 0x10].value
        SZ1 = dhdr[0x28, 0x11].value
    # image resolution
    SZ_VX2 = R[0]
    SZ_VX1 = R[1]

    # now sort the images along k-dimension
    k = np.argmin(abs(Orn[:3] + Orn[3:]))
    # sorted indeces
    si = np.argsort(P[:, k])
    Pos = np.zeros(P.shape, dtype=np.float64)
    im = np.zeros((SZ0, SZ1, SZ2), dtype=np.float32)

    # check if the detentions are in agreement (the pixel array could be transposed...)
    if IM[0].shape[0] == SZ1:
        for i in range(SZ0):
            im[i, :, :] = IM[si[i]] * SI[si[i], 0] + SI[si[i], 1]
            Pos[i, :] = P[si[i]]
    else:
        for i in range(SZ0):
            im[i, :, :] = IM[si[i]].T * SI[si[i], 0] + SI[si[i], 1]
            Pos[i, :] = P[si[i]]

    # proper slice thickness
    Zz = (P[si[-1], 2] - P[si[0], 2]) / (SZ0-1)
    Zy = (P[si[-1], 1] - P[si[0], 1]) / (SZ0-1)
    Zx = (P[si[-1], 0] - P[si[0], 0]) / (SZ0-1)

    # dictionary for affine and image size for the image
    A = {
        'AFFINE': np.array([[SZ_VX2 * Orn[0], SZ_VX1 * Orn[3], Zx, Pos[0, 0]],
                            [SZ_VX2 * Orn[1], SZ_VX1 * Orn[4], Zy, Pos[0, 1]],
                            [SZ_VX2 * Orn[2], SZ_VX1 * Orn[5], Zz, Pos[0, 2]], [0., 0., 0., 1.]]),
        'SHAPE': (SZ0, SZ1, SZ2)}

    # the returned image is already scaled according to the dcm header
    return {'im': im, 'affine': A['AFFINE'], 'shape': A['SHAPE'], 'orient': ornt, 'sruid': sruid}



"""
Compile CUDA source code and setup Python 3 package 'nimpa'
for namespace 'niftypet'.
"""
import logging
import re
import sys
from pathlib import Path

from setuptools import find_packages, setup
from setuptools_scm import get_version

from niftypet.ninst import cudasetup as cs
from niftypet.ninst import install_tools as tls

__version__ = get_version(root=".", relative_to=__file__)

logging.basicConfig(level=logging.INFO, format=tls.LOG_FORMAT)
log = logging.getLogger("nimpa.setup")

tls.check_platform()
ext = tls.check_depends() # external dependencies

cs.resources_setup(gpu=False) # install resources.py
try:
    cs.dev_setup()            # update resources.py with a supported GPU device
except Exception as exc:
    log.error("could not set up CUDA:\n%s", exc)

# get the local path to NiftyPET resources.py
path_resources = cs.path_niftypet_local()
# if exists, import the resources and get the constants
resources = cs.get_resources()
# get the current setup, if any
Cnt = resources.get_setup()

build_ver = ".".join(__version__.split('.')[:3]).split(".dev")[0]
# some setup kwargs cannot be in `pyproject.toml` since
# `install_requires` is dynamically set depending on CUDA GPU detection
setup_kwargs = {
    "use_scm_version": True, "packages": find_packages(exclude=["tests"]), "package_data": {
        "niftypet": [
            "nimpa/auxdata/*", "nimpa/acr_design/core_mumap/*", "nimpa/acr_design/core_nac/*",
            "nimpa/acr_design/rods/*", "nimpa/acr_design/sampling/*"]}, "install_requires": [
                'dcm2niix', 'dipy>=1.3.0', 'imageio', 'miutil[nii]>=0.10.0', 'nibabel>=2.4.0',
                'ninst>=0.12.0', 'numpy>=1.14', 'pydicom>=1.0.2', 'scipy', 'setuptools', 'spm12',
                'SimpleITK']}
# 'SimpleITK>=1.2.0'
cmake_args = [
    f"-DNIMPA_BUILD_VERSION={build_ver}",
    f"-DNIMPA_KERNEL_RADIUS={getattr(resources, 'RSZ_PSF_KRNL', 8)}"]

try:
    import cuvec as cu
    from miutil import cuinfo
    from skbuild import setup as sksetup
    assert cu.include_path.is_dir()
    try:
        nvcc_arch_raw = map(cuinfo.compute_capability, range(cuinfo.num_devices()))
        nvcc_arches = {"%d%d" % i for i in nvcc_arch_raw if i >= (3, 5)}
        if nvcc_arches:
            cmake_args.append("-DCMAKE_CUDA_ARCHITECTURES=" + ";".join(sorted(nvcc_arches)))
    except Exception as exc:
        if "sdist" not in sys.argv or any(i in sys.argv for i in ["build", "bdist", "wheel"]):
            log.warning("CUDA device detection error:\n%s", exc)
            log.warning("Compiling for all architectures")
except Exception as exc:
    log.warning("Import or CUDA device detection error:\n%s", exc)
    setup(**setup_kwargs)
else:
    setup_kwargs['install_requires'].extend(["cuvec>=2.3.1", "numcu"])
    for i in (Path(__file__).resolve().parent / "_skbuild").rglob("CMakeCache.txt"):
        i.write_text(re.sub("^//.*$\n^[^#].*pip-build-env.*$", "", i.read_text(), flags=re.M))
    sksetup(cmake_source_dir="niftypet", cmake_languages=("C", "CXX", "CUDA"),
            cmake_minimum_required_version="3.18", cmake_args=cmake_args, **setup_kwargs)
